/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "scheme.h"
#ifdef FILES_HAVE_FDS
#include <sys/types.h>
#include <sys/time.h>
#endif
#ifndef NO_USER_BREAK_HANDLER
#include <signal.h>
#endif

void mzsleep(int v, void *fds)
{
  if (v) {
#ifndef NO_SLEEP
    sleep(v);
#endif
  } else {
#ifdef FILES_HAVE_FDS
    int limit;
    fd_set rd, ex;
    
#ifdef USE_ULIMIT
    limit = ulimit(4, 0);
#else
    limit = getdtablesize();
#endif    

    memcpy(&rd, fds, sizeof(fd_set));
    memcpy(&ex, fds, sizeof(fd_set));
    select(limit, &rd, NULL, &ex, NULL);
#endif
  }
}

#ifndef NO_USER_BREAK_HANDLER

static int break_flag;
static Scheme_Object *(*orig_evaluator)(Scheme_Object *);

void user_break_hit()
{
  break_flag = 1;
}

int check_break_flag()
{
  if (break_flag) {
    break_flag = 0;
    return 1;
  } else
    return 0;
}

Scheme_Object *adjust_break_flag_and_eval(Scheme_Object *expr)
{
  break_flag = 0;
  return orig_evaluator(expr);
}

#endif

#define GDESC "Globals and symbols are case-sensitive.\n"
#define KDESC "Builtin globals are constant.\n"
#define SDESC "Set! works on undefined symbols.\n"
#define EDESC "Call/cc is replaced with call/ec.\n"
#define NDESC "Keywords not enforced.\n"

int main(int argc, char *argv[])
{
  Scheme_Env *global_env;
  char *prog, **evals_and_loads, *real_switch = NULL;
  int *is_load, num_enl;
  int i;
  int no_more_switches = 0;
  int script_mode = 0;
  int no_rep = 0;
  Scheme_Object *sch_argv;

  scheme_sleep = mzsleep;

  prog = argv[0];
  argv++;
  --argc;

  evals_and_loads = (char **)malloc(sizeof(char *) * argc);
  is_load = (int *)malloc(sizeof(int) * argc);
  num_enl = 0;

  while (!no_more_switches && argc && argv[0][0] == '-') {
    real_switch = argv[0];

    if (!strcmp("--case-sens", argv[0]))
      argv[0] = "-g";
    else if (!strcmp("--const-globs", argv[0]))
      argv[0] = "-k";
    else if (!strcmp("--esc-cont", argv[0]))
      argv[0] = "-c";
    else if (!strcmp("--set-undef", argv[0]))
      argv[0] = "-s";
    else if (!strcmp("--no-key", argv[0]))
      argv[0] = "-n";
    else if (!strcmp("--script", argv[0]))
      argv[0] = "-r";

    if (!argv[0][1] || (argv[0][1] == '-' && argv[0][2])) {
      goto bad_switch;
    } else {
      char *str;
      for (str = argv[0] + 1; *str; str++) {
	switch (*str) {
	case 'g':
	  scheme_case_sensitive = 1;
	  break;
	case 'c':
	  scheme_escape_continuations_only = 1;
	  break;
	case 'k':
	  scheme_constant_builtins = 1;
	  break;
	case 's':
	  scheme_allow_set_undefined = 1;
	  break;
	case 'n':
	  scheme_no_keywords = 1;
	  break;
	case 'e':
	  if (argc < 2) {
	    printf("%s: No expression following %s switch\n", 
		   prog, real_switch);
	    return -1;
	  }
	  argv++;
	  --argc;
	  evals_and_loads[num_enl] = argv[0];
	  is_load[num_enl++] = 0;
	  break;
	case 'r':
	  script_mode = 1;
	  no_more_switches = 1;
	case 'f':
	  if (argc < 2) {
	    printf("%s: No file name following %s switch\n", 
		   prog, real_switch);
	    return -1;
	  }
	  argv++;
	  --argc;
	  evals_and_loads[num_enl] = argv[0];
	  is_load[num_enl++] = 1;
	  break;
	case 'v':
	  no_rep = 1;
	  break;
	case '-':
	  no_more_switches = 1;
	  break;
	default:
	  goto bad_switch;
	}
      }
    }
    argv++;
    --argc;
  }

  if (!script_mode) {
    printf(scheme_banner());
    if (scheme_case_sensitive)
      printf(GDESC);
    if (scheme_escape_continuations_only)
      printf(EDESC);
    if (scheme_constant_builtins)
      printf(KDESC);
    if (scheme_allow_set_undefined)
      printf(SDESC);
    if (scheme_no_keywords)
      printf(NDESC);
  }
  
  global_env = scheme_basic_env();

  sch_argv = scheme_make_vector(argc, scheme_null);
  for (i = 0; i < argc; i++)
    SCHEME_VEC_ELS(sch_argv)[i] = scheme_make_string(argv[i]);

  scheme_add_global("argv", sch_argv, global_env);

  for (i = 0; i < num_enl; i++) {
    if (is_load[i]) {
      if (!scheme_load(evals_and_loads[i]))
	break;
    } else if (!scheme_setjmp(scheme_error_buf))
      scheme_eval_string(evals_and_loads[i], global_env);
    else
      break;
  }
  
  if (!no_rep && !script_mode) {
#ifndef NO_USER_BREAK_HANDLER
    signal(SIGINT, user_break_hit);
    scheme_check_for_break = check_break_flag;
    scheme_enable_break = 1;
    
    orig_evaluator = scheme_evaluator;
    scheme_evaluator = adjust_break_flag_and_eval;
#endif

    /* enter read-eval-print loop */
    scheme_rep();
    printf ("\nbye\n");
  }

  return 0;

 bad_switch:
  printf("%s: bad switch %s\n", prog, real_switch);
  printf("Supported switches:\n"
	 "  -g, --case-sens : " GDESC
	 "  -c, --esc-cont : " EDESC
	 "  -k, --const-globs: " KDESC
	 "  -s, --set-undef : " SDESC
	 "  -n, --no-key : " NDESC
	 "  -r, --script : Script mode: use as last flag for scripts\n" 
	 "  -e <expr> : Evaluates <expr> after MzScheme starts.\n"
	 "  -f <file> : Loads <file> after MzScheme starts.\n"
	 "  -v : Suppresses the read-eval-print loop.\n"
	 "  -- : No argument following this switch is used as a switch.\n"
	 "All remaining arguments are put into the MzScheme variable argv.\n"
         "Expressions and files are evaluated/loaded in the specified order.\n");
  return -1;
}
