/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#ifdef USE_STACKAVAIL
#include <malloc.h>
#endif
#ifdef UNIX_FIND_STACK_BOUNDS
#include <signal.h>
#include <sys/time.h>
#include <sys/resource.h>
#endif
#include "schmach.h"

/* globals */

Scheme_Object *scheme_eval_waiting;
Scheme_Object *scheme_eval_wait_expr;
Scheme_Object *scheme_multiple_values;
Scheme_Env *scheme_eval_wait_env;
int stack_grows_up;
int scheme_fuel_counter;

/* locals */
static Scheme_Object *eval (int argc, Scheme_Object *argv[]);
static Scheme_Object *compile (int argc, Scheme_Object *argv[]);
static Scheme_Object *expand(int argc, Scheme_Object **argv);
static Scheme_Object *expand_once(int argc, Scheme_Object **argv);
/* static Scheme_Object *expand_annotate(int argc, Scheme_Object **argv); */
static Scheme_Object *enable_break (int argc, Scheme_Object *argv[]);
static Scheme_Object *user_break_poll(int argc, Scheme_Object *argv[]);
static Scheme_Object *dynamic_enable_break (int argc, Scheme_Object *argv[]);
static Scheme_Object *do_default_eval_handler(int argc, Scheme_Object **argv);
static Scheme_Object *current_eval(int argc, Scheme_Object **argv);

static Scheme_Object *write_application(Scheme_Object *obj);
static Scheme_Object *read_application(Scheme_Object *obj);
static Scheme_Object *write_sequence(Scheme_Object *obj);
static Scheme_Object *read_sequence(Scheme_Object *obj);
static Scheme_Object *write_branch(Scheme_Object *obj);
static Scheme_Object *read_branch(Scheme_Object *obj);
static Scheme_Object *write_syntax(Scheme_Object *obj);
static Scheme_Object *read_syntax(Scheme_Object *obj);

static Scheme_Object *define_symbol, *letrecs_symbol, *lambda_symbol;
static Scheme_Object *unknown_symbol, *void_symbol, *quote_symbol;
static Scheme_Object *defmacro_symbol, *letmacro_symbol, *begin_symbol;
static Scheme_Object *define_struct_symbol, *let_struct_symbol;

static Scheme_Object *global_variable_symbol;
static Scheme_Object *local_variable_symbol;
static Scheme_Object *local_variable_indirect_symbol;
static Scheme_Object *constant_value_symbol;

static Scheme_Debugging_Info *eval_debug = NULL;

static Scheme_Object *default_eval_handler;

int scheme_in_macro_application = 0;

#define cons(x,y) scheme_make_pair(x,y)

typedef struct {
  Scheme_Type type;
  short num_args;
  Scheme_Object *args[1];
  /* After array of f & args, array of chars for eval type */
} Scheme_App_Rec;

void
scheme_init_eval (Scheme_Env *env)
{
  if (scheme_starting_up) {
    scheme_eval_waiting = scheme_alloc_eternal_object();
    scheme_eval_waiting->type = scheme_eval_waiting_type;

    scheme_multiple_values = scheme_alloc_eternal_object();
    scheme_multiple_values->type = scheme_multiple_values_type;

    define_symbol = scheme_intern_symbol("define");
    defmacro_symbol = scheme_intern_symbol("defmacro");
    letrecs_symbol = scheme_intern_symbol("letrec*");
    lambda_symbol = scheme_intern_symbol("lambda");
    unknown_symbol = scheme_intern_symbol("unknown");
    void_symbol = scheme_intern_symbol("void");
    quote_symbol = scheme_intern_symbol("quote");
    letmacro_symbol = scheme_intern_symbol("letmacro");
    begin_symbol = scheme_intern_symbol("begin");
    define_struct_symbol = scheme_intern_symbol("define-struct");
    let_struct_symbol = scheme_intern_symbol("let-struct");

    global_variable_symbol = scheme_intern_symbol("global-variable");
    local_variable_symbol = scheme_intern_symbol("local-variable");
    local_variable_indirect_symbol 
      = scheme_intern_symbol("local-variable-indirect");
    constant_value_symbol = scheme_intern_symbol("constant-value");

    default_eval_handler = scheme_make_prim_w_arity(do_default_eval_handler,
						    1, 1);

    scheme_install_type_writer(scheme_application_type, write_application);
    scheme_install_type_reader(scheme_application_type, read_application);
    scheme_install_type_writer(scheme_sequence_compilation_type, write_sequence);
    scheme_install_type_reader(scheme_sequence_compilation_type, read_sequence);
    scheme_install_type_writer(scheme_branch_compilation_type, write_branch);
    scheme_install_type_reader(scheme_branch_compilation_type, read_branch);
    scheme_install_type_writer(scheme_syntax_compilation_type, write_syntax);
    scheme_install_type_reader(scheme_syntax_compilation_type, read_syntax);
  }

  scheme_add_global_constant("eval", 
			     scheme_make_prim_w_arity(eval, 1, 1), 
			     env);
  scheme_add_global_constant("compile", 
			     scheme_make_prim_w_arity(compile, 1, 1), 
			     env);
  scheme_add_global_constant("expand", 
			     scheme_make_prim_w_arity(expand, 1, 1), 
			     env);
  scheme_add_global_constant("expand-once", 
			     scheme_make_prim_w_arity(expand_once, 1, 1), 
			     env);
#if 0
  scheme_add_global_constant("expand-and-annotate", 
			     scheme_make_prim_w_arity(expand_annotate, 1, 1), 
			     env);
#endif
  scheme_add_global_constant("enable-break", 
			     scheme_make_prim_w_arity(enable_break, 0, 1), 
			     env);
  scheme_add_global_constant("dynamic-enable-break", 
			     scheme_make_prim_w_arity(dynamic_enable_break, 
						      1, 1), 
			     env);
  scheme_add_global_constant("user-break-poll", 
			     scheme_make_prim_w_arity(user_break_poll, 0, 1), 
			     env);
  scheme_add_global_constant("current-eval",
			     scheme_make_prim_w_arity(current_eval, 0, 1),
			     env);
}

Scheme_Object *scheme_debug_get_last_eval(void)
{
  Scheme_Object *obj;

  obj = scheme_alloc_object();
  obj->type = scheme_debug_handle_type;
  SCHEME_PTR_VAL(obj) = eval_debug;

  return obj;
}

/* 4 cases => magic number for some compilers doing a switch */
enum {
  SCHEME_EVAL_CONSTANT = 0,
  SCHEME_EVAL_GLOBAL,
  SCHEME_EVAL_LOCAL,
  SCHEME_EVAL_GENERAL
};

static int eval_type(Scheme_Object *obj)
{
  Scheme_Type type;

  type = SCHEME_TYPE(obj);

  if (type > _scheme_values_types_)
    return SCHEME_EVAL_CONSTANT;
  else if (SAME_TYPE(type, scheme_static_distance_type))
    return SCHEME_EVAL_LOCAL;
  else if (SAME_TYPE(type, scheme_static_distance_unbox_type))
    return SCHEME_EVAL_LOCAL;
  else if (SAME_TYPE(type, scheme_variable_type))
    return SCHEME_EVAL_GLOBAL;
  else
    return SCHEME_EVAL_GENERAL;
}    

static Scheme_Object *scheme_make_application(Scheme_Object *v,
					      Scheme_Object *src)
{
  Scheme_Object *o;
  Scheme_App_Rec *app;
  char *evals;
  int i, n;

  o = v;
  n = 0;
  while (!SCHEME_NULLP(o)) {
    n++;
    if (!SCHEME_LISTP(o))
      scheme_signal_error("wrong expression form discovered");
    o = SCHEME_CDR(o);
  }

  app = (Scheme_App_Rec *)scheme_malloc_stubborn(sizeof(Scheme_App_Rec)
						 + ((n - 1) 
						    * sizeof(Scheme_Object *))
						 + n * sizeof(char));

  app->num_args = n - 1;

  evals = ((char *)app) + sizeof(Scheme_App_Rec) 
    + (app->num_args * sizeof(Scheme_Object *));

  app->args[0] = SCHEME_CAR(v);
  evals[0] = eval_type(app->args[0]);

  v = SCHEME_CDR(v);
  for (i = 1; i < n; i++, v = SCHEME_CDR(v)) {
    app->args[i] = SCHEME_CAR(v);
    evals[i] = eval_type(app->args[i]);
  }

  app->type = scheme_application_type;

  scheme_end_stubborn_change((void *)app);

  return (Scheme_Object *)app;
}

Scheme_Object *scheme_make_syntax_compilation(Scheme_Syntax_Compilation *prim,
					      Scheme_Object *data)
{
  Scheme_Object *v;

  v = scheme_alloc_stubborn_object();
  v->type = scheme_syntax_compilation_type;
  SCHEME_PTR1_VAL(v) = (void *)prim;
  SCHEME_PTR2_VAL(v) = (void *)data;
  scheme_end_stubborn_change((void *)v);

  return v;
}

void scheme_init_compile_recs(Scheme_Compile_Info *src, 
			      Scheme_Compile_Info *dest, int n)
{
  int i;

  for (i = 0; i < n; i++) {
    dest[i].can_optimize_constants = src->can_optimize_constants;
    dest[i].base_let_depth = src->base_let_depth;
  }
}

void scheme_merge_compile_recs(Scheme_Compile_Info *src, 
			       Scheme_Compile_Info *dest, int n)
{
  int i;

  if (!n) {
    src->can_return_closure = 0;
    src->max_let_depth = 0;
    return;
  }
  
  src->can_return_closure = dest[0].can_return_closure;
  src->max_let_depth = dest[0].max_let_depth;

  for (i = 1; i < n; i++) {
    if (dest[i].max_let_depth > src->max_let_depth)
      src->max_let_depth = dest[i].max_let_depth;
    if (dest[i].can_return_closure)
      src->can_return_closure = 1;
  }
}

Scheme_Object *
scheme_compile_list(Scheme_Object *form, Scheme_Env *env, 
		    Scheme_Compile_Info *rec)
{
  if (SCHEME_NULLP(form)) {
    rec->can_return_closure = 0;
    rec->max_let_depth = 0;
    return scheme_null;
  } else if (SCHEME_PAIRP(form)) {
    Scheme_Compile_Info recs[2];
    Scheme_Object *c1, *c2;

    scheme_init_compile_recs(rec, recs, 2);

    c1 = scheme_compile_expr(SCHEME_CAR(form), env, &recs[0]);
    c2 = scheme_compile_list(SCHEME_CDR(form), env, &recs[1]);

    scheme_merge_compile_recs(rec, recs, 2);

    return scheme_make_pair(c1, c2);
  } else
    return scheme_compile_expr(form, env, rec);
}

Scheme_Object *scheme_compile(Scheme_Object *form, Scheme_Env *env)
{
  Scheme_Compile_Info rec;

  rec.base_let_depth = -1;
  rec.can_optimize_constants = 1;
  return scheme_compile_expr(form, env, &rec);
}

Scheme_Object *scheme_compile_simple(Scheme_Object *form, Scheme_Env *env,
				     Scheme_Compile_Info *rec_in)
{
  Scheme_Compile_Info rec;

  rec.base_let_depth = -1;
  rec.can_optimize_constants = rec_in->can_optimize_constants;
  return scheme_compile_expr(form, env, &rec);
}

static Scheme_Object *
scheme_compile_expand_expr(Scheme_Object *form, Scheme_Env *env, 
			   Scheme_Compile_Info *rec, int depth)
{
  Scheme_Object *name, *var, *rest, *val;

  if (rec) {
    rec->max_let_depth = 0;
    rec->can_return_closure = 0;
  }

  if (!SCHEME_PAIRP(form)) {
    if (SCHEME_SYMBOLP(form)) {
      var = scheme_static_distance(form, env, 
				   rec ? rec->can_optimize_constants : 0);
      if (SAME_TYPE(SCHEME_TYPE(var), scheme_variable_type))
	val = (SCHEME_VAR_BUCKET(var))->val;
      else
	val = var;
      if (val) {
	if (SAME_TYPE(SCHEME_TYPE(val), scheme_macro_type))
	  scheme_signal_error("illegal use of macro name \"%s\"",
			      SCHEME_SYM_VAL(form));
	else if (SAME_TYPE(SCHEME_TYPE(val), scheme_compiled_syntax_type))
	  scheme_signal_error("illegal use of syntax name %s "
			      "(local binding missing?)",
			      SCHEME_SYM_VAL(form));
      }

      if (rec)
	return var;
      else if (depth < 0) {
	/* Expand & annotate! */
	if (SAME_TYPE(SCHEME_TYPE(var), scheme_variable_type))
	  return cons(global_variable_symbol,
		      cons(form, scheme_null));
	else {
	  int j;
	  j = SCHEME_SDFRAME_VAL(var);
	  while (j--)
	    env = env->basic.next;
	  if (env->basic.flags & SCHEME_HAS_ANNOTATIONS)
	    form = env->values[env->basic.num_bindings + SCHEME_SDPOS_VAL(var)];
	  return cons((env->basic.autounbox 
		       ? local_variable_indirect_symbol : local_variable_symbol),
		      cons(form, 
			   scheme_null));
	}
      }
    } else if (rec && SAME_TYPE(SCHEME_TYPE(form), scheme_closure_type))
      return scheme_make_closure_compilation(SCHEME_CLOS_ENV(form),
					     SCHEME_CLOS_CODE(form), 
					     rec);

    if (depth < 0)
      return cons(constant_value_symbol,
		  cons(form, scheme_null));

    return form;
  }

  name = SCHEME_CAR(form);
  rest = SCHEME_CDR(form);
  if (SCHEME_SYMBOLP(name)) {
    /* Check for macros: */
    var = scheme_static_distance(name, env, 
				 rec ? rec->can_optimize_constants : 0);
    if (SAME_TYPE(SCHEME_TYPE(var), scheme_static_distance_type)) {
      /* apply to local variable: compile away! */
    } else {
      if (SAME_TYPE(SCHEME_TYPE(var), scheme_variable_type))
	val = (SCHEME_VAR_BUCKET(var))->val;
      else
	val = var;

      if (val) {
	if (SAME_TYPE(SCHEME_TYPE(val), scheme_macro_type)) {
	  /* Macro expression: expand and compile */
	  if (!depth)
	    return form; /* We've gone as deep as requested */
	  scheme_in_macro_application++;
	  form = scheme_apply_to_list((Scheme_Object *)SCHEME_PTR_VAL(val), 
				      rest);
	  --scheme_in_macro_application;
	  if (rec)
	    return scheme_compile_expr(form, env, rec);
	  else {
	    --depth;
	    if (depth)
	      return scheme_expand_expr(form, env, depth);
	    else
	      return form;
	  }
	} else if (SAME_TYPE(SCHEME_TYPE(val), scheme_compiled_syntax_type)) {
	  if (rec) {
	    Scheme_Syntax *f;
	    f = SCHEME_SYNTAX(val);
	    return f(form, env, rec);
	  } else {
	    Scheme_Syntax_Expander *f;
	    f = SCHEME_SYNTAX_EXP(val);
	    return f(form, env, depth);
	  }
	}
	/* Else: other global - must be a function: compile away! */
      }
      /* Else: unknown global - must be a function: compile away! */

      /* Optimize (void) to just the value void */
      if (rec && rec->can_optimize_constants && SAME_OBJ(var, scheme_void_func)
	  && SCHEME_NULLP(rest))
	return scheme_void;
    }
    
    if (rec) {
      var = scheme_make_pair(var, scheme_compile_list(rest, env, rec));
      return scheme_make_application(var, form);
    } else
      return scheme_expand_list(form, env, depth);
  } else if (rec)
    return scheme_make_application(scheme_compile_list(form, env, rec), form);
  else
    return scheme_expand_list(form, env, depth);
}

Scheme_Object *scheme_compile_expr(Scheme_Object *form, Scheme_Env *env, 
				   Scheme_Compile_Info *rec)
{
  return scheme_compile_expand_expr(form, env, rec, 1);
}

Scheme_Object *scheme_expand_expr(Scheme_Object *form, Scheme_Env *env, 
				  int depth)
{
  return scheme_compile_expand_expr(form, env, NULL, depth);
}

static Scheme_Object *check_immediate_macro(Scheme_Object *first, 
					    Scheme_Env *env, 
					    Scheme_Compile_Info *rec, 
					    int depth,
					    Scheme_Object **current_val)
{
  Scheme_Object *name, *val, *var;

 check_top:
  *current_val = NULL;

  name = SCHEME_CAR(first);
  if (!SCHEME_SYMBOLP(name))
    return first;

  var = scheme_static_distance(name, env, 0);
  if (!SAME_TYPE(SCHEME_TYPE(var), scheme_variable_type))
    return first;

  val = (SCHEME_VAR_BUCKET(var))->val;
  if (!val)
    return first;

  *current_val = val;

  if (!SAME_TYPE(SCHEME_TYPE(val), scheme_macro_type))
    return first;

  /* Yep, it's a macro; expand once */
  first = scheme_expand_expr(first, env, 1);

  if (SCHEME_PAIRP(first))
    goto check_top;
  
  return first;
}

static Scheme_Object *
scheme_compile_expand_block(Scheme_Object *forms, Scheme_Env *env, 
			    Scheme_Compile_Info *rec, int depth)
/* This ugly code parses a block of code, transforming embedded
  define, defmacro, and define-struct into letrec*, letmacro, and
  let-struct. It is espcailly ugly because we have to expand macros
  before deciding what we have. */
{
  Scheme_Object *first;
  Scheme_Compile_Info recs[2];

  if (rec) {
    rec->max_let_depth = 0;
    rec->can_return_closure = 0;
  }

  if (SCHEME_NULLP(forms))
    return scheme_null;

  first = SCHEME_CAR(forms);

  if (SCHEME_PAIRP(first)) {
    Scheme_Object *name, *gval;

    /* Check for macro expansion, which could mask the real
       define, defmacro, or define-struct: */
    first = check_immediate_macro(first, env, rec, depth, &gval);

    name = SCHEME_CAR(first);
    if (SAME_OBJ(gval, scheme_define_syntax)) {
      /* Turn defines into a letrec*: */
      Scheme_Object *var, *v, *link, *l = scheme_null, *start = NULL;
      
      while (1) {
	v = SCHEME_CDR(first);
	
	if (!SCHEME_PAIRP(v))
	  scheme_signal_error("define (embedded): bad form (improper list)");

	if (SCHEME_PAIRP(v) &&
	    SCHEME_PAIRP(SCHEME_CAR(v))) {
	  /* Transform to a lambda */
	  var = SCHEME_CAR(SCHEME_CAR(v));
	  v = cons(var,
		   cons(cons(lambda_symbol,
			     cons(SCHEME_CDR(SCHEME_CAR(v)),
				  SCHEME_CDR(v))),
			scheme_null));
	} else
	  var = SCHEME_CAR(v);
	
	if (!SCHEME_SYMBOLP(var))
	  scheme_signal_error("define (embedded): "
			      "variable name must be a symbol");

	link = scheme_make_pair(v, scheme_null);
	if (!start)
	  start = link;
	else
	  SCHEME_CDR(l) = link;
	l = link;
	forms = SCHEME_CDR(forms);
	SCHEME_ASSERT(SCHEME_LISTP(forms), "bad embedded `define'");

	/* Special case: (define define x) */
	if (SAME_OBJ(define_symbol, var))
	  break;

	if (!SCHEME_NULLP(forms)) {
	  first = SCHEME_CAR(forms);
	  if (SCHEME_PAIRP(first)) {
	    first = check_immediate_macro(first, env, rec, depth, &gval);
	    name = SCHEME_CAR(first);
	    if (NOT_SAME_OBJ(gval, scheme_define_syntax))
	      break;
	  } else
	    break;
	} else
	  break;
      }
      
      forms = scheme_make_pair(letrecs_symbol, scheme_make_pair(start, forms));
      
      name = NULL;
    } else if (SAME_OBJ(gval, scheme_defmacro_syntax)) {
      /* Convert to letmacro */
       Scheme_Object *var, *params, *body, *rest;

       rest = SCHEME_CDR(forms);

       first = SCHEME_CDR(first);
       if (!SCHEME_PAIRP(first))
	 scheme_signal_error("bad embedded `defmacro' form");
       var = SCHEME_CAR(first);
       first = SCHEME_CDR(first);
       if (!SCHEME_PAIRP(first))
	 scheme_signal_error("bad embedded `defmacro' form");
       params = SCHEME_CAR(first);
       body = SCHEME_CDR(first);
       
       forms = cons(letmacro_symbol,
		    cons(var,
			 cons(params,
			      cons(cons(begin_symbol, body),
				   rest))));

       name = NULL;
    } else if (SAME_OBJ(gval, scheme_define_struct_syntax)) {
      /* Convert to let-struct */
      Scheme_Object *fields, *rest;

      rest = SCHEME_CDR(forms);

      first = SCHEME_CDR(first);
      if (!SCHEME_PAIRP(first))
	scheme_signal_error("bad embedded `define-struct' form");
      name = SCHEME_CAR(first);
      first = SCHEME_CDR(first);
      if (!SCHEME_PAIRP(first))
	scheme_signal_error("bad embedded `define-struct' form");
      fields = SCHEME_CAR(first);
      first = SCHEME_CDR(first);
      if (!SCHEME_NULLP(first))
	scheme_signal_error("bad embedded `define-struct' form");

      forms = cons(let_struct_symbol,
		   cons(name,
			cons(fields,
			     rest)));

      name = NULL;
    }

    if (!name) {
      if (rec)
	forms = scheme_compile_expr(forms, env, rec);
      else {
	--depth;
	if (depth)
	  forms = scheme_expand_expr(forms, env, depth);
      }
      
      return scheme_make_pair(forms, scheme_null);
    }
  }

  if (rec) {
    scheme_init_compile_recs(rec, recs, 2);

    first = scheme_compile_expr(first, env, &recs[0]);
    forms = scheme_compile_expand_block(SCHEME_CDR(forms), env, &recs[1], 1);

    scheme_merge_compile_recs(rec, recs, 2);
  } else {
    first = scheme_expand_expr(first, env, depth);
    forms = scheme_compile_expand_block(SCHEME_CDR(forms), env, rec, depth);
  }

  return scheme_make_pair(first, forms);
}

Scheme_Object *
scheme_compile_block(Scheme_Object *forms, Scheme_Env *env, 
		     Scheme_Compile_Info *rec)
{
  return scheme_compile_expand_block(forms, env, rec, 1);
}

Scheme_Object *
scheme_expand_block(Scheme_Object *forms, Scheme_Env *env, 
		    int depth)
{
  return scheme_compile_expand_block(forms, env, NULL, depth);
}

Scheme_Object *
scheme_expand_list(Scheme_Object *form, Scheme_Env *env, int depth)
{
  if (SCHEME_NULLP(form))
    return scheme_null;

  if (!SCHEME_PAIRP(form))
    scheme_signal_error("bad syntax in expand");

  return scheme_make_pair(scheme_expand_expr(SCHEME_CAR(form), env, depth),
			  scheme_expand_list(SCHEME_CDR(form), env, depth));
}

Scheme_Object *
scheme_expand_simple(Scheme_Object *form, Scheme_Env *env, int depth)
{
  return cons(SCHEME_CAR(form),
	      scheme_expand_list(SCHEME_CDR(form), env, depth));
}

typedef struct {
  Scheme_Type type;
  Scheme_Object *test;
  Scheme_Object *tbranch;
  Scheme_Object *fbranch;
} Scheme_Branch_Rec;

Scheme_Object *
scheme_make_branch_compilation(Scheme_Object *test, Scheme_Object *thenp,
			       Scheme_Object *elsep)
{
  Scheme_Branch_Rec *b;

  b = (Scheme_Branch_Rec *)scheme_malloc_stubborn(sizeof(Scheme_Branch_Rec));
  b->type = scheme_branch_compilation_type;

  /* Try to partial evaluate: (if (not x) y z) => (if x z y) */
  /* This only happens if built-in globals are constant */
  if (SAME_TYPE(SCHEME_TYPE(test), scheme_application_type)) {
    Scheme_App_Rec *app;

    app = (Scheme_App_Rec *)test;
    if ((app->num_args == 1) && SAME_PTR(scheme_not_prim, app->args[0])) {
      test = thenp;
      thenp = elsep;
      elsep = test;
      test = app->args[1];
    }
  }


  b->test = test;
  b->tbranch = thenp;
  b->fbranch = elsep;

  scheme_end_stubborn_change((void *)b);

  return (Scheme_Object *)b;
}

Scheme_Object * scheme_make_sequence_compilation(Scheme_Object *seq)
{
  Scheme_Object **array;
  Scheme_Sequence *o;
  int count, i;

  count = scheme_list_length(seq);

  if (!count)
    return scheme_compiled_void(0);

  o = scheme_malloc_stubborn(sizeof(Scheme_Sequence) +
			     (count - 1) * sizeof(Scheme_Object *));
  o->type = scheme_sequence_compilation_type;
  o->count = count;
  array = o->array;
  
  for (i = 0; i < count; i++) {
    array[i] = SCHEME_CAR(seq);
    seq = SCHEME_CDR(seq);
  }

  scheme_end_stubborn_change(o);

  return (Scheme_Object *)o;
}

#if MAX_QUICK_VALUES > 0
Scheme_Env_Stack *scheme_alloc_env_stack(Scheme_Env_Stack *prev)
{
  Scheme_Env_Stack *env_stack;

  env_stack = scheme_malloc(sizeof(Scheme_Env_Stack));
  env_stack->stack_pos = NUM_QUICK_ENVS_PER_SEGMENT - 2;
  env_stack->prev = prev;
  env_stack->next = NULL;

  return env_stack;
}
#endif

#ifndef STACK_SAFETY_MARGIN
#define STACK_SAFETY_MARGIN 50000
#endif

#ifndef REALLY_BIG_STACK
#define REALLY_BIG_STACK 10000000
#endif

#ifdef UNIX_FIND_STACK_BOUNDS
static jmp_buf stack_init_jmp_buf;
unsigned long stack_boundary;

#ifdef SEARCH_FIND_STACK_LIMIT
void stack_fault_handler(int sig)
{
  longjmp(stack_init_jmp_buf, 1);
}
#endif
#endif

unsigned long get_deeper_address()
{
  int v, *vp;
  vp = &v;
  return (unsigned long)vp;
}

 /* OBSOLETE: */
 /* REALLY_BIG_STACK <X> sets the "really big" threadhold for a stack
     size to <X>. It is used with UNIX_FIND_STACK_BOUNDS. If the stack 
     is reportedly at least this big, then no stack crawl will occur
     to find a precise boundary. Instead, the stack limit will be
     approximated as 90% of the stack size beyond the current
     stack position. If REALLY_BIG_STACK is unspecified, it defaults
     to 10MB. To turn off REALLY_BIG_STACK, set it to 0. */

#ifdef UNIX_FIND_STACK_BOUNDS
#ifndef SEARCH_FIND_STACK_LIMIT
extern unsigned long GC_get_stack_base();
#endif
#endif

void scheme_init_stack_check()
{
  int *v;
  unsigned long deeper;
#ifdef UNIX_FIND_STACK_BOUNDS
  struct rlimit rl;
#endif
  
  deeper = get_deeper_address();
  stack_grows_up = (deeper > (unsigned long)&v);

#ifdef STACK_GROWS_UP
  if (!stack_grows_up) {
    if (scheme_console_printf)
      scheme_console_printf("Stack grows DOWN, not UP.\n");
    else
      printf("Stack grows DOWN, not UP.\n");
    exit(-1);
  }
#endif
#ifdef STACK_GROWS_DOWN
  if (stack_grows_up) {
    if (scheme_console_printf)
      scheme_console_printf("Stack grows UP, not DOWN.\n");
    else
      printf("Stack grows UP, not DOWN.\n");
    exit(-1);
  }
#endif


#ifdef UNIX_FIND_STACK_BOUNDS
  getrlimit(RLIMIT_STACK, &rl);

#ifndef SEARCH_FIND_STACK_LIMIT

  stack_boundary = GC_get_stack_base();
  if (stack_grows_up)
    stack_boundary += ((unsigned long)rl.rlim_cur - STACK_SAFETY_MARGIN);
  else
    stack_boundary += (STACK_SAFETY_MARGIN - (unsigned long)rl.rlim_cur);

#else

  /* OBSOLETE: */
#if REALLY_BIG_STACK
  if (rl.rlim_cur > REALLY_BIG_STACK) {
    unsigned long amount = (unsigned long)(0.9 * rl.rlim_cur);

    stack_boundary = ((unsigned long)&v);

    if (stack_grows_up)
      stack_boundary += amount;
    else
      stack_boundary -= amount;

    return;
  }
#endif

  if (rl.rlim_cur >= RLIM_INFINITY) {
    if (stack_grows_up)
      stack_boundary = (unsigned long)0xFFFFFFFF;
    else
      stack_boundary = 0;
  } else {
    static void (*old_segv_handler)(int), (*old_bus_handler)(int);
    unsigned long definite_boundary;
    unsigned long stepsize = 100;
    int toofar = 0;
    int *old_v;
    static int h;

    old_segv_handler = signal(SIGSEGV, stack_fault_handler);
    old_bus_handler = signal(SIGBUS, stack_fault_handler);
    
    v = (int *)&v;
    
    definite_boundary = (unsigned long)v;
    if (stack_grows_up)
      definite_boundary += rl.rlim_cur;
    else
      definite_boundary -= rl.rlim_cur;

    while (1) {
      old_v = v;

      if (stack_grows_up)
	v += stepsize;
      else
	v -= stepsize;

      if ((stack_grows_up && (unsigned long)v < (unsigned long)old_v)
	  || (!stack_grows_up && (unsigned long)v > (unsigned long)old_v)) {
	/* Numerical overflow */
	v = old_v;
      } else if ((stack_grows_up 
		  && ((unsigned long)v >= definite_boundary))
		 || (!stack_grows_up 
		     && ((unsigned long)v <= definite_boundary))){
	/* Too far by current & rl.rlim_cur */
	toofar = 1;
	v = old_v;
      } else {
	if (!setjmp(stack_init_jmp_buf))
	  h = *v;
	else {
	  /* Address overflow */
	  toofar = 1;
	  signal(SIGSEGV, stack_fault_handler);
	  signal(SIGBUS, stack_fault_handler);
	  v = old_v;
	}
      }
      if (toofar) {
	stepsize = stepsize / 2;
	if (stepsize < 100)
	  break;
      } else
	stepsize *= 2;
    }
    
    (void)signal(SIGSEGV, old_segv_handler);
    (void)signal(SIGBUS, old_bus_handler);
    
    stack_boundary = (unsigned long)old_v;
    if (stack_grows_up)
      stack_boundary -= STACK_SAFETY_MARGIN;
    else
      stack_boundary += STACK_SAFETY_MARGIN;
  }
#endif
  /* END OBSOLETE */
#endif
}

#define MAX_QUICK_RANDS 5
static Scheme_Object *zero_rands_ptr;

unsigned long previous;

Scheme_Object *
scheme_do_eval (Scheme_Object *obj, int num_rands, Scheme_Env *env, 
		int get_value)
{
  Scheme_Type type;
  Scheme_Object *v, **rands;
#if MAX_QUICK_RANDS > 0
  Scheme_Object *quick_rands[MAX_QUICK_RANDS];
#endif
#if MAX_QUICK_VALUES > 0
  int use_quick = -1;
  int old_stack_pos;
  Scheme_Env_Stack *env_stack;
#endif

#ifdef DO_STACK_CHECK
#ifdef USE_STACKAVAIL
  if (stackavail() < STACK_SAFETY_MARGIN)
#else
#ifdef UNIX_FIND_STACK_BOUNDS 
  unsigned long pos;

  pos = (unsigned long)&pos;

  if (STK_COMP(pos, stack_boundary))
#endif
#endif
    {
#ifndef ERROR_ON_OVERFLOW
    scheme_tail_rator = obj;
    scheme_tail_num_rands = num_rands;
    scheme_tail_rands = (Scheme_Object **)env;
    scheme_init_jmpup_buf(&scheme_overflow_cont);
    if (scheme_setjmpup(&scheme_overflow_cont, 
			scheme_current_process->cc_start)) {
      scheme_init_jmpup_buf(&scheme_overflow_cont);
      if (!scheme_overflow_reply) {
	scheme_longjmp(scheme_error_buf, 1);
      } else
	return scheme_overflow_reply;
    } else
      scheme_longjmp(scheme_current_process->overflow_buf, 1);
#else
    if (!scheme_current_process->stack_overflow) {
      scheme_current_process->stack_overflow = 1;
      scheme_signal_error("out of stack space (infinite loop?)");
    }
#endif
  }
#endif

  previous = pos;

#define DO_CHECK_FOR_BREAK() \
	if ((scheme_fuel_counter--) <= 0) { \
	  scheme_process_block(0); \
	  scheme_fuel_counter = scheme_current_process->engine_weight; \
	}

  if (num_rands >= 0) {
    rands = (Scheme_Object **)env;

  apply_top:

    type = SCHEME_TYPE (obj);
    if (type == scheme_prim_type)
      v = SCHEME_PRIM(obj)(num_rands, rands);
    else if (type == scheme_compiled_closure_type)
      {
	Scheme_Closure_Compilation_Data *data;
	Scheme_Env *frame;
	int i, n, has_rest, num_params;

	DO_CHECK_FOR_BREAK();

	data = (Scheme_Closure_Compilation_Data *)SCHEME_CLOS_CODE(obj);
	env = SCHEME_CLOS_ENV(obj);
	obj = data->code;

	num_params = data->num_params;
	has_rest = data->has_rest;

	n = num_params + data->extra_frame_slots;

	if (n) {
#if MAX_QUICK_VALUES > 0
	  if (!data->must_allocate && (n <= MAX_QUICK_VALUES)) {
	    if (use_quick < 0) {
	      env_stack = scheme_current_process->env_stack;
	      old_stack_pos = env_stack->stack_pos;
	      if (old_stack_pos)
		use_quick = --env_stack->stack_pos;
	      else {
		if (!env_stack->next)
		  env_stack->next = scheme_alloc_env_stack(env_stack);
		env_stack = scheme_current_process->env_stack 
		  = env_stack->next;
		env_stack->stack_pos = NUM_QUICK_ENVS_PER_SEGMENT - 1;
		use_quick = NUM_QUICK_ENVS_PER_SEGMENT - 1;
	      }
	    }
	    frame = (Scheme_Env *)(env_stack->array + use_quick);
	    
	    frame->basic.autounbox = 0;
	  } else
#endif
	    frame = scheme_new_frame(n);

	  env = scheme_extend_env (frame, env);
	}

	if (num_params) {
	  if (has_rest) {
	    if (num_rands < (num_params - 1)) {
	      scheme_signal_error ("too few arguments to procedure"
				   " (need at least %d, provided %d)", 
				   num_params - 1, num_rands);
	      return scheme_void; /* Doesn't get here */
	    }
	  } else if (num_rands != num_params) {
	    if (num_rands < num_params) {
	      scheme_signal_error ("too few arguments to procedure"
				   " (need %d, provided %d)", 
				   num_params, num_rands);
	      return scheme_void; /* Doesn't get here */
	    } else {
	      scheme_signal_error ("too many arguments to procedure"
				   " (%d allowed, provided %d)",
				   num_params, num_rands);
	      return scheme_void; /* Doesn't get here */
	    }
	  }
	

	  n = num_params;
	  if (has_rest)
	    --n;

	  for (i = 0; i < n; i++)
	    scheme_add_good_binding(i, rands[i], frame);

	  if (has_rest) {
	    Scheme_Object *rest_vals, *pairs;
	    int extra;

	    extra = num_rands - n;
	    rest_vals = scheme_null;
	    if (extra) {
	      for (i = num_rands - 1; i >= n; --i) {
		pairs = scheme_alloc_object();
		pairs->type = scheme_pair_type;
		SCHEME_CDR(pairs) = rest_vals;
		SCHEME_CAR(pairs) = rands[i];
		rest_vals = pairs;
		pairs++;
	      }
	    }

	    scheme_add_good_binding (n, rest_vals, frame);
	  }
	} else if (num_rands) {
	  scheme_signal_error ("procedure expects 0 arguments"
			       " (provided %d)", num_rands);
	  return scheme_void; /* Doesn't get here */
	}

	goto eval_top;
      }
    else switch (type) {
    case scheme_closed_prim_type:
      v = SCHEME_CLSD_PRIM(obj)(SCHEME_CLSD_PRIM_DATA(obj),
				num_rands, rands);
      break;
    case scheme_cont_type:
      {
	Scheme_Cont *c;
	Scheme_Dynamic_Wind *dw, *common;
	Scheme_Object *value;

	DO_CHECK_FOR_BREAK();

	if (num_rands != 1) 
	  value = scheme_values(num_rands, rands);
	else
	  value = rands[0];

	c = (Scheme_Cont *)obj;
	SCHEME_ASSERT(*c->ok,
		      "apply: attempted to cross continuation boundary");
	SCHEME_ASSERT(SAME_OBJ(c->home, scheme_current_process),
		      "apply: attempted to apply foreign continuation");
	
	common = scheme_current_process->dw;
	while (common) {
	  dw = c->dw;
	  while (dw && dw != common)
	    dw = dw->prev;
	  if (dw)
	    break;
	  common = common->prev;
	}
	
	c->common = common;
	for (dw = scheme_current_process->dw; dw != common; dw = dw->prev)
	  if (dw->post)
	    dw->post(dw->data);
	
	if (!common && scheme_current_process->dw) {
	  /* Reset jump buf from first DW record */
	  for (dw = scheme_current_process->dw; dw->prev; dw = dw->prev);
	  memcpy(&scheme_error_buf, &dw->saveerr, sizeof(jmp_buf));
	}
	
	c->value = value;
	scheme_longjmpup(&c->buf);
    }
    case scheme_struct_proc_type:
      v = scheme_apply_struct_proc(obj, num_rands, rands);
      break;
    case scheme_escaping_cont_type:
      {
	Scheme_Object *value;

	if (num_rands != 1) 
	  value = scheme_values(num_rands, rands);
	else
	  value = rands[0];

	SCHEME_ASSERT(SCHEME_CONT_HOME(obj),
		      "apply: attempt to jump into an escape continuation");
	SCHEME_ASSERT(SAME_OBJ(SCHEME_CONT_HOME(obj), scheme_current_process),
		      "apply: attempted to apply foreign escape continuation");
	SCHEME_CONT_VAL(obj) = value;
	scheme_jumping_to_continuation = 1;
	scheme_longjmp(scheme_error_buf, 1);
      }
    case scheme_case_closure_type:
      {
	Scheme_Sequence *seq;
	Scheme_Closure_Compilation_Data *data;

	int i;

	seq = (Scheme_Sequence *)obj;
	for (i = 0; i < seq->count; i++) {
	  data = 
	    (Scheme_Closure_Compilation_Data *)SCHEME_CLOS_CODE(seq->array[i]);
	  if ((!data->has_rest && (data->num_params == num_rands))
	      || (data->has_rest && (data->num_params - 1 <= num_rands))) {
	    obj = seq->array[i];
	    goto apply_top;
	  }
	}

	scheme_signal_error ("apply: no case in case-lambda matching %d args",
			     num_rands);
      }
    default:
      scheme_signal_error ("apply: bad procedure (type was %s)", 
			   scheme_get_type_name(type));
      v = scheme_null;
      break;
    }
  } else {

  eval_top:

    type = SCHEME_TYPE (obj);
    switch (type)
      {
      case scheme_variable_type:
	{
#define global_lookup(prefix, _obj, tmp)                            \
	  tmp = (SCHEME_VAR_BUCKET(_obj))->val;                     \
	  if (!tmp)                                                 \
	    scheme_signal_error ("reference to unbound symbol: %s", \
		 SCHEME_SYM_VAL((SCHEME_VAR_BUCKET(_obj))->key));   \
	  prefix tmp

	  global_lookup(v = , obj, v);  
	  goto returnv;
	}
      case scheme_static_distance_type:
	{
	  int j;

#define local_lookup(prefix, _obj, env, senv)                   \
	  j = SCHEME_SDFRAME_VAL(_obj);                         \
          if (j) {                                              \
            env = senv->basic.next;                             \
	    while (--j)                                         \
	      env = env->basic.next;                            \
	    v = env->values[SCHEME_SDPOS_VAL(_obj)];            \
            if (env->basic.autounbox) v = SCHEME_PTR_VAL(v);    \
	  } else {                                              \
            v = senv->values[SCHEME_SDPOS_VAL(_obj)];           \
            if (senv->basic.autounbox) v = SCHEME_PTR_VAL(v);   \
	  }                                                     \
	  prefix (v)

	  local_lookup(v =, obj, env, env);
	  goto returnv;
	}
      case scheme_static_distance_unbox_type:
	{
	  int j;

	  local_lookup(v = , obj, env, env);
	  goto returnv;
	}
      case scheme_syntax_compilation_type:
	{
	  Scheme_Syntax_Compilation *f;
	  
	  f = (Scheme_Syntax_Compilation *)SCHEME_PTR1_VAL(obj);
	  v = f((Scheme_Object *)SCHEME_PTR2_VAL(obj), env);
	  break;
	}
      case scheme_application_type:
	{
	  Scheme_App_Rec *app;
	  Scheme_Object **args, *tmpv, **randsp;
	  char *evals;
	  int j, k;
	  Scheme_Env *frame;

	  app = (Scheme_App_Rec *)obj;
	  num_rands = app->num_args;
	  
	  evals = ((char *)obj) + sizeof(Scheme_App_Rec)
	    + (num_rands * sizeof(Scheme_Object *));

	  obj = app->args[0];
	  
	  /* Inline local & global variable lookups for speed */
	  switch (evals[0]) {
	  case SCHEME_EVAL_CONSTANT:
	    break;
	  case SCHEME_EVAL_GLOBAL:
	    global_lookup(obj =, obj, v);
	    break;
	  case SCHEME_EVAL_LOCAL:
	    local_lookup(obj =, obj, frame, env);
	    break;
	  case SCHEME_EVAL_GENERAL:
	    obj = _scheme_eval_compiled(obj, env);
	    break;
	  }
	  
	  if (num_rands) {
#if MAX_QUICK_RANDS > 0
	    if (get_value && (num_rands <= MAX_QUICK_RANDS))
	      rands = quick_rands;
	    else
#endif
	      rands = (Scheme_Object **)scheme_malloc(sizeof(Scheme_Object *)
						      * num_rands);
	    
	    /* Inline local & global variable lookups for speed */
	    evals++;
	    args = app->args + 1;
	    randsp = rands;
	    for (k = num_rands; k--; ) {
	      v = *(args++);
	      switch (*(evals++)) {
	      case SCHEME_EVAL_CONSTANT:
		*(randsp++) = v;
		break;
	      case SCHEME_EVAL_GLOBAL:
		global_lookup(*(randsp++) =, v, tmpv);
		break;
	      case SCHEME_EVAL_LOCAL:
		local_lookup(*(randsp++) =, v, frame, env);
		break;
	      case SCHEME_EVAL_GENERAL:
		*(randsp++) = _scheme_eval_compiled(v, env);
		break;
	      }
	    }
	  } else
	    rands = &zero_rands_ptr;
	  
	  if (get_value)
	    goto apply_top;
	  else {
	    /* inlined tail call "invocation": */
	    scheme_tail_rator = obj;
	    scheme_tail_num_rands = num_rands;
	    scheme_tail_rands = rands;
	    
	    v = scheme_tail_call_waiting;
	    goto returnv;
	  }
	}
      
      case scheme_sequence_compilation_type:
	{
	  Scheme_Object **array;
	  int i;
	  
	  i = ((Scheme_Sequence *)obj)->count;
	  array = ((Scheme_Sequence *)obj)->array;
	  
	  while (--i)
	    _scheme_eval_compiled(*(array++), env);

	  if (get_value) {
	    obj = *array;
	    goto eval_top;
	  } else {
	    scheme_eval_wait_expr = *array;
	    scheme_eval_wait_env = env;
	    v = scheme_eval_waiting;
	    goto returnv;
	  }
	}
      case scheme_branch_compilation_type:
	{
	  Scheme_Branch_Rec *b;
	  b = (Scheme_Branch_Rec *)obj;
	  obj = NOT_SAME_OBJ(_scheme_eval_compiled(b->test, env), scheme_false)
	    ? b->tbranch : b->fbranch;

	  goto eval_top;
	}
      case scheme_closure_compilation_type:
	v = scheme_make_compiled_closure(env, obj);
	goto returnv;

      case scheme_let_value_compilation_type:
	{
	  Scheme_Let_Value *lv;
	  Scheme_Object *value;
	  int i;

	  lv = (Scheme_Let_Value *)obj;
	  i = lv->position;
	  value = lv->value;
	  obj = lv->body;

	  value = _scheme_eval_compiled(value, env);
	  scheme_add_good_binding(i, value, env);

	  goto eval_top;
	}

      case scheme_let_void_compilation_type:
	{
	  Scheme_Let_Void *lv;
	  int start, end, i;

	  lv = (Scheme_Let_Void *)obj;
	  start = lv->start;
	  end = lv->end;
	  obj = lv->body;

	  for (i = start; i < end; i++)
	    scheme_add_good_binding(i, scheme_void, env);

	  goto eval_top;
	}

      default:
	v = obj;
	goto returnv;
      }
  }

  if (get_value) {
    if (SAME_OBJ(v, scheme_tail_call_waiting)) {
      obj = scheme_tail_rator;
      num_rands = scheme_tail_num_rands;
      rands = scheme_tail_rands;
      goto apply_top;
    }

    if (SAME_OBJ(v, scheme_eval_waiting)) {
      obj = scheme_eval_wait_expr;
      env = scheme_eval_wait_env;
      goto eval_top;
    }
  }

 returnv:
  if (get_value > 0)
    if (SAME_OBJ(v, scheme_multiple_values))
      scheme_signal_error("returned %d values to single-value context",
			  scheme_multiple_count);

#if MAX_QUICK_VALUES > 0
  if (use_quick >= 0) {
    if (!old_stack_pos) {
      env_stack->next = NULL;
      scheme_current_process->env_stack = env_stack->prev;
    } else
      env_stack->stack_pos = old_stack_pos;
  }
#endif

  return v;
}

Scheme_Object *scheme_eval (Scheme_Object *obj, Scheme_Env *env)
{
  return scheme_eval_compiled(scheme_compile(obj, env), env);
}

Scheme_Object *scheme_eval_compiled(Scheme_Object *obj, Scheme_Env *env)
{
  return scheme_top_level_eval(obj, -1, (Scheme_Object **)env);
}

Scheme_Object *scheme_expand (Scheme_Object *obj, Scheme_Env *env)
{
  return scheme_expand_expr(obj, env, 0x7FFF);
}

Scheme_Object *scheme_tail_eval_expr(Scheme_Object *obj, Scheme_Env *env)
{
  return scheme_tail_eval(obj, env);
}

Scheme_Object *scheme_standard_evaluator(Scheme_Object *expr)
{
  if (scheme_eval_handler)
    return _scheme_apply(scheme_eval_handler, 1, &expr);
  else
    return _scheme_eval_compiled(scheme_compile(expr, scheme_env), 
				 scheme_env);
}

/* local functions */

static Scheme_Object *
eval (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("eval", 1, 1, argc, argv);

  return scheme_evaluator(argv[0]);
}

static Scheme_Object *
do_default_eval_handler(int argc, Scheme_Object **argv)
{
  if (argc != 1)
    scheme_wrong_count("default-eval-handler", 1, 1, argc, argv);

  return scheme_evaluator(argv[0]);
}

static Scheme_Object *
current_eval(int argc, Scheme_Object **argv)
{
  if (argc > 1)
    scheme_wrong_count("eval-handler", 0, 1, argc, argv);
  
  if (!argc) {
    if (scheme_eval_handler)
      return scheme_eval_handler;
    return default_eval_handler;
  }

  if (!SCHEME_PROCP(argv[0]))
    scheme_wrong_type("eval-handler", "procedure", 0, argc, argv);

  scheme_eval_handler = argv[0];

  return scheme_void;
}

static Scheme_Object *
compile (int argc, Scheme_Object *argv[])
{
  Scheme_Compile_Info rec;

  if (argc != 1)
    scheme_wrong_count("compile", 1, 1, argc, argv);
  
  rec.base_let_depth = -1;
  rec.can_optimize_constants = 0;
  return scheme_compile_expr(argv[0], scheme_env, &rec);
}

static Scheme_Object *
expand(int argc, Scheme_Object **argv)
{
  if (argc != 1)
    scheme_wrong_count("expand", 1, 1, argc, argv);

  return scheme_expand_expr(argv[0], scheme_env, 0x7FFF);
}

static Scheme_Object *
expand_once(int argc, Scheme_Object **argv)
{
  if (argc != 1)
    scheme_wrong_count("expand-once", 1, 1, argc, argv);

  return scheme_expand_expr(argv[0], scheme_env, 1);
}

#if 0
static Scheme_Object *
expand_annotate(int argc, Scheme_Object **argv)
{
  if (argc != 1)
    scheme_wrong_count("expand-once", 1, 1, argc, argv);

  return scheme_expand_expr(argv[0], scheme_env, -1);
}
#endif

Scheme_Object *scheme_eval_string(const char *str, Scheme_Env *env)
{
  Scheme_Object *port, *expr;

  port = scheme_make_string_input_port(str);
  expr = scheme_read(port);
  return scheme_eval(expr, env);
}

Scheme_Object *enable_break (int argc, Scheme_Object *argv[])
{
  if (argc > 1)
    scheme_wrong_count("enable-break", 0, 1, argc, argv);
  
  if (argc) {
    scheme_enable_break = NOT_SAME_OBJ(argv[0], scheme_false);
    return scheme_void;
  } else
    return scheme_enable_break ? scheme_true : scheme_false;
}

typedef struct {
  int old;
  Scheme_Object *f;
} Enable_Break;

static void pre_enable_br(void *eb)
{
  ((Enable_Break *)eb)->old = scheme_enable_break;
  scheme_enable_break = 1;
}

static void post_enable_br(void *eb)
{
  scheme_enable_break = ((Enable_Break *)eb)->old;
}

static Scheme_Object *do_enable_br(void *eb)
{
  return _scheme_apply(((Enable_Break *)eb)->f, 0, NULL);
}

Scheme_Object *dynamic_enable_break (int argc, Scheme_Object *argv[])
{
  Enable_Break *eb;

  if (argc != 1)
    scheme_wrong_count("dynamic-enable-break", 1, 1, argc, argv);
  
  eb = (Enable_Break *)scheme_malloc(sizeof(Enable_Break));
  eb->f = argv[0];

  return scheme_dynamic_wind(pre_enable_br, do_enable_br, post_enable_br,
			     NULL, eb);
}

static Scheme_Object *default_user_break_poll_proc(int c, Scheme_Object **a)
{
  if (c)
    scheme_wrong_count("default-user-break-poll-proc", 1, 1, c, a);

  if (scheme_check_for_break)
    return scheme_check_for_break() ? scheme_true : scheme_false;
  else
    return scheme_false;
}

static Scheme_Object *user_break_poll(int argc, Scheme_Object *argv[])
{
  if (argc > 1)
    scheme_wrong_count("user-break-poll", 0, 1, argc, argv);

  if (!argc) {
    if (scheme_current_process->user_break_poll_proc)
      return scheme_current_process->user_break_poll_proc;
    else
      return scheme_make_prim_w_arity(default_user_break_poll_proc, 1, 1);
  } else {
    if (!SCHEME_PROCP(argv[0]))
      scheme_wrong_type("user-break-poll", "closure", 0, argc, argv);
    scheme_current_process->user_break_poll_proc = argv[0];
    return scheme_void;
  }
}

/****************************************************/

#define BOOL(x) (x ? scheme_true : scheme_false)

static Scheme_Object *write_application(Scheme_Object *obj)
{
  Scheme_App_Rec *app;
  Scheme_Object *l;
  int i;

  app = (Scheme_App_Rec *)obj;

  l = scheme_null;
  for (i = app->num_args + 1; i--; )
    l = cons(app->args[i], l);
  
  return l;
}

static Scheme_Object *read_application(Scheme_Object *obj)
{
  /* (void) optimization: */
  if (SCHEME_PAIRP(obj) && SCHEME_NULLP(SCHEME_CDR(obj))
      && SAME_OBJ(SCHEME_CAR(obj), scheme_void_func))
    return scheme_void;

  return scheme_make_application(obj, scheme_void);
}

static Scheme_Object *write_sequence(Scheme_Object *obj)
{
  Scheme_Object *l, **a;
  int i;

  i = ((Scheme_Sequence *)obj)->count;
  a = ((Scheme_Sequence *)obj)->array;

  l = scheme_null;
  for (; i--; )
    l = cons(a[i], l);
  
  return l;
}

static Scheme_Object *read_sequence(Scheme_Object *obj)
{
  return scheme_make_sequence_compilation(obj);
}

static Scheme_Object *write_branch(Scheme_Object *obj)
{
  Scheme_Branch_Rec *b;

  b = (Scheme_Branch_Rec *)obj;

  return cons(b->test, cons(b->tbranch, b->fbranch));
}

static Scheme_Object *read_branch(Scheme_Object *obj)
{
  if (!SCHEME_PAIRP(obj) || !SCHEME_PAIRP(SCHEME_CDR(obj)))
    scheme_signal_error("bad compiled branch");

  return scheme_make_branch_compilation(SCHEME_CAR(obj), 
					SCHEME_CAR(SCHEME_CDR(obj)),
					SCHEME_CDR(SCHEME_CDR(obj)));
}

static Scheme_Object *write_syntax(Scheme_Object *obj)
{
  Scheme_Object *sym;
  Scheme_Syntax_Compilation *f;

  f = (Scheme_Syntax_Compilation *)SCHEME_PTR1_VAL(obj);
  sym = scheme_find_executer_name(f);
  if (!sym)
    sym = unknown_symbol;

  return cons(sym, (Scheme_Object *)SCHEME_PTR2_VAL(obj));
}

static Scheme_Object *read_syntax(Scheme_Object *obj)
{
  Scheme_Object *sym;
  Scheme_Syntax_Compilation *f;

  if (!SCHEME_PAIRP(obj) || !SCHEME_SYMBOLP(SCHEME_CAR(obj)))
    scheme_signal_error("bad compiled syntax");

  sym = SCHEME_CAR(obj);
  if (SAME_OBJ(unknown_symbol, sym))
    scheme_signal_error("unknown (at write time) compiled syntax");

  if (SAME_OBJ(void_symbol, sym))
    return scheme_void;

  f = scheme_find_executer(sym);
  if (!f)
    scheme_signal_error("unknown compiled syntax");

  return scheme_make_syntax_compilation(f, SCHEME_CDR(obj));
}
