/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#include <string.h>
#include <ctype.h>
#include <memory.h>

#ifdef DOS_MEMORY
#include <dos.h>

#define PTR_TO_LONG(p) ((FP_SEG(p) << 4) + FP_OFF(p))
#else
#define PTR_TO_LONG(p) ((long)(p))
#endif

#define FILL_FACTOR 2

#define FOREVER_OK 1

Scheme_Hash_Table *
scheme_hash_table (int size, int type, int has_const, int forever)
{
  Scheme_Hash_Table *table;
  size_t asize;

  table = (Scheme_Hash_Table*) scheme_malloc (sizeof (Scheme_Hash_Table));
  table->size = size;
  table->step = size / 10;
  table->count = 0;

  table->type = scheme_hash_table_type;

  asize = (size_t)size * sizeof(Scheme_Bucket *);
#if FOREVER_OK
  if (forever) {
    table->buckets = (Scheme_Bucket **)scheme_malloc_atomic(asize);
    memset((char *)table->buckets, 0, asize);
  } else
#endif
    table->buckets = (Scheme_Bucket **)scheme_malloc(asize);

  table->by_address = ((type == SCHEME_hash_stringptr) ||
		       (type == SCHEME_hash_ptr));
  table->by_symbol = (type == SCHEME_hash_symbol);
  table->has_constants = has_const;
  table->forever = forever;

  return (table);
}

typedef int hash_v_t;

Scheme_Bucket *
get_bucket (Scheme_Hash_Table *table, const char *key, int add, 
	    Scheme_Bucket *b)
{
  hash_v_t h;
  Scheme_Bucket *bucket;

  if (table->by_address) {
    h = (PTR_TO_LONG(key) >> 2) % table->size;
  } else {
    int i;
    h = i = 0;
    while (key[i]) {
      int c = key[i++];
      if (!scheme_case_sensitive)
	if (isupper(c))
	  c = tolower(c);
      h += (h << 5) + h + c;
    }
    h = h % table->size;
  }
  if (h < 0) h = -h;

  while ((bucket = table->buckets[h])) {
    if ((table->by_symbol && !strcmp(key, SCHEME_SYM_VAL(bucket)))
	|| (!table->by_symbol 
	    && ((table->by_address && SAME_PTR(bucket->key, key))
		|| (!table->by_address && !strcmp(bucket->key, key)))))
      return bucket;
    h = (h + table->step) % table->size;
  }

  if (!add)
    return NULL;

  if (table->count * FILL_FACTOR >= table->size) {
    /* Rehash */
    int i, oldsize = table->size;
    size_t asize;
    Scheme_Bucket **old = table->buckets;

    table->size = (table->size * 2) + 113;

    asize = (size_t)table->size * sizeof(Scheme_Bucket *);
#if FOREVER_OK
    if (table->forever) {
      table->buckets = (Scheme_Bucket **)scheme_malloc_atomic(asize);
      memset((char *)table->buckets, 0, asize);
    } else
#endif
      table->buckets = (Scheme_Bucket **)scheme_malloc(asize);

    table->count = 0;
    for (i = 0; i < oldsize; i++)
      if (old[i])
	get_bucket(table, 
		   table->by_symbol ? SCHEME_SYM_VAL(old[i]) : old[i]->key, 
		   1, old[i]);
  }

  if (b) {
    bucket = b;
  } else {
    size_t bsize;

    if (!table->has_constants)
      bsize = sizeof(Scheme_Bucket);
    else
      bsize = sizeof(Scheme_Bucket_With_Const_Flag);

    if (table->forever)
      bucket = (Scheme_Bucket *)scheme_malloc_uncollectable(bsize);
    else
      bucket = (Scheme_Bucket *)scheme_malloc(bsize);

    bucket->type = scheme_variable_type;

    if (table->has_constants) {
      ((Scheme_Bucket_With_Const_Flag *)bucket)->const_val = 0;
      ((Scheme_Bucket_With_Const_Flag *)bucket)->keyword = 0;
    }

    bucket->key = (char *)key;
    bucket->val = NULL;
  }

  table->buckets[h] = bucket;

  table->count++;

  return bucket;
}

Scheme_Bucket *
scheme_bucket_from_table (Scheme_Hash_Table *table, const char *key)
{
  return get_bucket(table, key, 1, NULL);
}

void 
scheme_add_to_table (Scheme_Hash_Table *table, const char *key, void *val, 
		     int constant)
{
  Scheme_Bucket *b;

  b = get_bucket(table, key, 1, table->by_symbol ? (Scheme_Bucket *)val : NULL);

  if (table->by_symbol)
    return;

  if (table->has_constants
      && ((Scheme_Bucket_With_Const_Flag *)b)->const_val
      && b->val)
    scheme_signal_error("define: cannot redefine constant %s", 
			 SCHEME_SYM_VAL(key));

  if (val)
    b->val = val;
  if (constant && table->has_constants)
    ((Scheme_Bucket_With_Const_Flag *)b)->const_val = constant;
}

void *
scheme_lookup_in_table (Scheme_Hash_Table *table, const char *key)
{
  Scheme_Bucket *bucket = get_bucket(table, key, 0, NULL);

  if (bucket) {
    if (table->by_symbol)
      return (void *)bucket;
    else
      return bucket->val;
  } else
    return NULL;
}

void
scheme_change_in_table (Scheme_Hash_Table *table, const char *key, void *naya)
{
  Scheme_Bucket *bucket = get_bucket(table, key, 0, NULL);

  if (bucket)
    bucket->val = naya;
}
