/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#include <sys/types.h>
#ifndef NO_STAT
#include <sys/stat.h>
#endif
#ifdef FILES_HAVE_FDS
#include <sys/time.h>
#endif
#ifdef DIR_FUNCTION
#ifdef DIR_INCLUDE
#include <dir.h>
#endif
#endif
#ifdef IO_INCLUDE
#include <io.h>
#endif
#ifdef UNISTD_INCLUDE
#include <unistd.h>
#endif

typedef struct Scheme_Indexed_String
{
  char *string;
  int size;
  int index;
} Scheme_Indexed_String;

typedef struct {
  FILE *f;
  int ungotten_count;
  int ungotten_allocated;
  char *ungotten;
} Scheme_Input_File;

/* globals */
Scheme_Object *scheme_eof;
Scheme_Object *scheme_orig_stdout_port;

Scheme_Object *(*scheme_make_stdin)(void) = NULL;
Scheme_Object *(*scheme_make_stdout)(void) = NULL;
Scheme_Object *(*scheme_make_stderr)(void) = NULL;

/* locals */
static Scheme_Object *scheme_file_input_port_type;
static Scheme_Object *scheme_string_input_port_type;
static Scheme_Object *scheme_file_output_port_type;
static Scheme_Object *scheme_string_output_port_type;
static Scheme_Object *scheme_user_input_port_type;
static Scheme_Object *scheme_user_output_port_type;
static Scheme_Object *pipe_read_port_type;
static Scheme_Object *pipe_write_port_type;

static Scheme_Object *text_symbol, *binary_symbol;
static Scheme_Object *append_symbol, *error_symbol;
static Scheme_Object *replace_symbol, *truncate_symbol;

static Scheme_Object *load_prim;

/* generic ports */

static Scheme_Object *scheme_make_eof (void);
static Scheme_Object *input_port_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *output_port_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *current_input_port (int argc, Scheme_Object *argv[]);
static Scheme_Object *current_output_port (int argc, Scheme_Object *argv[]);
static Scheme_Object *current_error_port (int argc, Scheme_Object *argv[]);
static Scheme_Object *process_input_port (int argc, Scheme_Object *argv[]);
static Scheme_Object *process_output_port (int argc, Scheme_Object *argv[]);
static Scheme_Object *process_error_port (int argc, Scheme_Object *argv[]);
static Scheme_Object *make_input_port (int argc, Scheme_Object *argv[]);
static Scheme_Object *make_output_port (int argc, Scheme_Object *argv[]);
static Scheme_Object *open_input_file (int argc, Scheme_Object *argv[]);
static Scheme_Object *open_output_file (int argc, Scheme_Object *argv[]);
static Scheme_Object *close_input_port (int argc, Scheme_Object *argv[]);
static Scheme_Object *close_output_port (int argc, Scheme_Object *argv[]);
static Scheme_Object *read_f (int argc, Scheme_Object *argv[]);
static Scheme_Object *read_char (int argc, Scheme_Object *argv[]);
static Scheme_Object *read_line (int argc, Scheme_Object *argv[]);
static Scheme_Object *peek_char (int argc, Scheme_Object *argv[]);
static Scheme_Object *eof_object_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *char_ready_p (int argc, Scheme_Object *argv[]);
static Scheme_Object *sch_write (int argc, Scheme_Object *argv[]);
static Scheme_Object *display (int argc, Scheme_Object *argv[]);
static Scheme_Object *newline (int argc, Scheme_Object *argv[]);
static Scheme_Object *write_char (int argc, Scheme_Object *argv[]);
static Scheme_Object *load (int argc, Scheme_Object *argv[]);
/* non-standard */
static Scheme_Object *flush_output (int argc, Scheme_Object *argv[]);
static Scheme_Object *file_position (int argc, Scheme_Object *argv[]);
static Scheme_Object *open_input_string (int argc, Scheme_Object *argv[]);
static Scheme_Object *open_output_string (int argc, Scheme_Object *argv[]);
static Scheme_Object *get_output_string (int argc, Scheme_Object *argv[]);
static Scheme_Object *sch_pipe(int argc, Scheme_Object **args);
static Scheme_Object *file_exists(int argc, Scheme_Object **argv);
static Scheme_Object *delete_file(int argc, Scheme_Object **argv);

static Scheme_Object *read_accept_compiled(int argc, Scheme_Object *argv[]);

#ifdef DIR_FUNCTION
static Scheme_Object *current_directory(int argc, Scheme_Object *argv[]);
#endif

#ifdef PROCESS_FUNCTION
static Scheme_Object *sch_process(int c, Scheme_Object *args[]);
static Scheme_Object *sch_system(int c, Scheme_Object *args[]);
static Scheme_Object *sch_process_star(int c, Scheme_Object *args[]);
static Scheme_Object *sch_system_star(int c, Scheme_Object *args[]);
#endif

void 
scheme_init_port (Scheme_Env *env)
{
  if (scheme_starting_up) {
    scheme_eof = scheme_make_eof ();

    scheme_string_input_port_type = scheme_make_port_type ("<string-input-port>");
    scheme_string_output_port_type = scheme_make_port_type ("<string-output-port>");

    scheme_file_input_port_type = scheme_make_port_type ("<file-input-port>");
    scheme_file_output_port_type = scheme_make_port_type ("<file-output-port>");

    scheme_user_input_port_type = scheme_make_port_type ("<user-input-port>");
    scheme_user_output_port_type = scheme_make_port_type ("<user-output-port>");

    pipe_read_port_type = scheme_make_port_type ("<pipe-input-port>");
    pipe_write_port_type = scheme_make_port_type ("<pipe-output-port>");

    if (scheme_make_stdin)
      scheme_stdin_port = scheme_make_stdin();
    else
      scheme_stdin_port = scheme_make_named_file_input_port(stdin, "STDIN");
    if (scheme_make_stdout)
      scheme_stdout_port = scheme_make_stdout();
    else
      scheme_stdout_port = scheme_make_file_output_port(stdout);
    if (scheme_make_stderr)
      scheme_stderr_port = scheme_make_stderr();
    else
      scheme_stderr_port = scheme_make_file_output_port(stderr);
    scheme_orig_stdout_port = scheme_stdout_port;

    scheme_current_process->in_port = scheme_stdin_port;
    scheme_current_process->out_port = scheme_stdout_port;
    scheme_current_process->err_port = scheme_stderr_port;

    scheme_current_process->read_port = NULL;
    scheme_current_process->write_port = NULL;
    scheme_current_process->readerr_port = NULL;

    text_symbol = scheme_intern_symbol("text");
    binary_symbol = scheme_intern_symbol("binary");
    append_symbol = scheme_intern_symbol("append");
    error_symbol = scheme_intern_symbol("error");
    replace_symbol = scheme_intern_symbol("replace");
    truncate_symbol = scheme_intern_symbol("truncate");

    load_prim = scheme_make_prim_w_arity(load, 1, 1);
  }

  scheme_add_global_constant ("eof", scheme_eof, env);

  scheme_add_global_constant ("input-port?", scheme_make_prim_w_arity(input_port_p, 1, 1), env);
  scheme_add_global_constant ("output-port?", scheme_make_prim_w_arity(output_port_p, 1, 1), env);

  scheme_add_global_constant ("current-input-port", scheme_make_prim_w_arity(current_input_port, 0, 1), env);
  scheme_add_global_constant ("current-output-port", scheme_make_prim_w_arity(current_output_port, 0, 1), env);
  scheme_add_global_constant ("current-error-port", scheme_make_prim_w_arity(current_error_port, 0, 1), env);

  scheme_add_global_constant ("process-input-port", scheme_make_prim_w_arity(process_input_port, 1, 1), env);
  scheme_add_global_constant ("process-output-port", scheme_make_prim_w_arity(process_output_port, 1, 1), env);
  scheme_add_global_constant ("process-error-port", scheme_make_prim_w_arity(process_error_port, 1, 1), env);

  scheme_add_global_constant ("open-input-file", scheme_make_prim_w_arity(open_input_file, 1, 2), env);
  scheme_add_global_constant ("open-input-string", scheme_make_prim_w_arity(open_input_string, 1, 1), env);
  scheme_add_global_constant ("open-output-file", scheme_make_prim_w_arity(open_output_file, 1, 3), env);
  scheme_add_global_constant ("open-output-string", scheme_make_prim_w_arity(open_output_string, 1, 1), env);
  scheme_add_global_constant ("get-output-string", scheme_make_prim_w_arity(get_output_string, 1, 1), env);
  scheme_add_global_constant ("close-input-port", scheme_make_prim_w_arity(close_input_port, 1, 1), env);
  scheme_add_global_constant ("close-output-port", scheme_make_prim_w_arity(close_output_port, 1, 1), env);
  scheme_add_global_constant ("make-input-port", scheme_make_prim_w_arity(make_input_port, 3, 3), env);
  scheme_add_global_constant ("make-output-port", scheme_make_prim_w_arity(make_output_port, 2, 2), env);
  scheme_add_global_constant ("read", scheme_make_prim_w_arity(read_f, 0, 1), env);
  scheme_add_global_constant ("read-char", scheme_make_prim_w_arity(read_char, 0, 1), env);
  scheme_add_global_constant ("read-line", scheme_make_prim_w_arity(read_line, 0, 1), env);
  scheme_add_global_constant ("peek-char", scheme_make_prim_w_arity(peek_char, 0, 1), env);
  scheme_add_global_constant ("eof-object?", scheme_make_prim_w_arity(eof_object_p, 1, 1), env);
  scheme_add_global_constant ("char-ready?", scheme_make_prim_w_arity(char_ready_p, 0, 1), env);
  scheme_add_global_constant ("write", scheme_make_prim_w_arity(sch_write, 1, 2), env);
  scheme_add_global_constant ("display", scheme_make_prim_w_arity(display, 1, 2), env);
  scheme_add_global_constant ("newline", scheme_make_prim_w_arity(newline, 0, 1), env);
  scheme_add_global_constant ("write-char", scheme_make_prim_w_arity(write_char, 1, 2), env);
  scheme_add_global_constant ("load", load_prim, env);

  scheme_add_global_constant ("flush-output", scheme_make_prim_w_arity(flush_output, 0, 1), env);
  scheme_add_global_constant ("file-position", scheme_make_prim_w_arity(file_position, 1, 2), env);

  scheme_add_global_constant ("pipe", scheme_make_prim_w_arity(sch_pipe, 0, 0), env);

  scheme_add_global_constant ("file-exists?", scheme_make_prim_w_arity(file_exists, 1, 1), env);
  scheme_add_global_constant ("delete-file", scheme_make_prim_w_arity(delete_file, 1, 1), env);

  scheme_add_global_constant ("read-accept-compiled", scheme_make_prim_w_arity(read_accept_compiled, 0, 1), env);

#ifdef DIR_FUNCTION
  scheme_add_global_constant ("current-directory", scheme_make_prim_w_arity(current_directory, 0, 1), env);
#endif

#ifdef PROCESS_FUNCTION
  scheme_add_global_constant ("process", scheme_make_prim_w_arity(sch_process, 1, 1), env);
  scheme_add_global_constant ("system", scheme_make_prim_w_arity(sch_system, 1, 1), env);
  scheme_add_global_constant ("process*", scheme_make_prim_w_arity(sch_process_star, 1, -1), env);
  scheme_add_global_constant ("system*", scheme_make_prim_w_arity(sch_system_star, 1, -1), env);
#endif
}

static Scheme_Object *
scheme_make_eof (void)
{
  Scheme_Object *eof;

  eof = scheme_alloc_eternal_object ();
  eof->type = scheme_eof_type;
  return (eof);
}

Scheme_Object *scheme_make_port_type(const char *name)
{
  return scheme_intern_symbol(name);
}

Scheme_Input_Port *
scheme_make_input_port (Scheme_Object *subtype,
			void *data,
			int (*getc_fun) (Scheme_Input_Port*),
			int (*char_ready_fun) (Scheme_Input_Port*),
			void (*close_fun) (Scheme_Input_Port*),
			void (*need_wakeup_fun)(Scheme_Input_Port*, void *))
{
  Scheme_Input_Port *ip;

  ip = (Scheme_Input_Port *) scheme_malloc (sizeof (Scheme_Input_Port));
  ip->type = scheme_input_port_type;
  ip->sub_type = subtype;
  ip->port_data = data;
  ip->getc_fun = getc_fun;
  ip->char_ready_fun = char_ready_fun;
  ip->need_wakeup_fun = need_wakeup_fun;
  ip->close_fun = close_fun;
  ip->name = "stdin";
  ip->ungotten = NULL;
  ip->ungotten_count = 0;
  ip->ungotten_allocated = 0;
  ip->position = 0;
  ip->lineNumber = 1;
  ip->charsSinceNewline = 0;
  ip->closed = 0;
  ip->eoffound = 0;
  return (ip);
}

Scheme_Output_Port *
scheme_make_output_port (Scheme_Object *subtype,
			 void *data,
			 void (*write_string_fun)(char *str, long,
						  Scheme_Output_Port*),
			 void (*close_fun) (Scheme_Output_Port*))
{
  Scheme_Output_Port *op;

  op = (Scheme_Output_Port *) scheme_malloc (sizeof (Scheme_Output_Port));
  op->type = scheme_output_port_type;
  op->sub_type = subtype;
  op->port_data = data;
  op->write_string_fun = write_string_fun;
  op->close_fun = close_fun;
  op->closed = 0;
  return (op);
}

#define check_closed(closed) if (closed) scheme_signal_error("port is closed")

int
scheme_getc (Scheme_Object *port)
{
  int c;
  Scheme_Input_Port *ip;
  
  ip = (Scheme_Input_Port *)port;
  check_closed(ip->closed);

  if (ip->ungotten_count)
    c = ip->ungotten[--ip->ungotten_count];
  else
    c = ip->getc_fun(ip);

  if (c != EOF) {
    ip->position++;
    if (c == '\n') {
      ip->charsSinceNewline = 1;
      ip->lineNumber++;
    } else
      ip->charsSinceNewline++;
  } else
    ip->eoffound = 1;

  return c;
}

void
scheme_ungetc (int ch, Scheme_Object *port)
{
  Scheme_Input_Port *ip;

  ip = (Scheme_Input_Port *)port;
  check_closed(ip->closed);
  if (ip->eoffound)
    return;

  if (ip->ungotten_count == ip->ungotten_allocated) {
    char *old;
    int oldc;

    old = ip->ungotten;
    oldc = ip->ungotten_count;
    if (oldc)
      ip->ungotten_allocated = 2 * oldc;
    else
      ip->ungotten_allocated = 5;

    ip->ungotten = (char *)scheme_malloc_atomic(ip->ungotten_allocated);
    if (oldc)
      memcpy(ip->ungotten, old, oldc);
  }
  ip->ungotten[ip->ungotten_count++] = ch;

  if (ip->position)
    --ip->position;
  if (!(--ip->charsSinceNewline)) {
    --ip->lineNumber;
    /* If you back up over two lines, then lineNumber will be wrong. */
  }
}

int
scheme_char_ready (Scheme_Object *port)
{
  Scheme_Input_Port *ip;

  ip = (Scheme_Input_Port *)port;
  check_closed(ip->closed);

  if (ip->ungotten_count)
    return 1;

  return ((ip->char_ready_fun) (ip));
}

void
scheme_need_wakeup (Scheme_Object *port, void *fds)
{
  Scheme_Input_Port *ip;

  ip = (Scheme_Input_Port *)port;
  
  if (ip->need_wakeup_fun)
    ((ip->need_wakeup_fun)(ip, fds));
}

long
scheme_tell (Scheme_Object *port)
{
  Scheme_Input_Port *ip;

  ip = (Scheme_Input_Port *)port;
  check_closed(ip->closed);

  return ip->position;
}

long
scheme_tell_line (Scheme_Object *port)
{
  Scheme_Input_Port *ip;

  ip = (Scheme_Input_Port *)port;
  check_closed(ip->closed);

  return ip->lineNumber;
}

void
scheme_close_input_port (Scheme_Object *port)
{
  Scheme_Input_Port *ip;

  ip = (Scheme_Input_Port *)port;
  if (ip->closed)
    return;

  if (ip->close_fun)
    ip->close_fun(ip);
  ip->closed = 1;
}

void
scheme_close_output_port (Scheme_Object *port)
{
  Scheme_Output_Port *op;

  op = (Scheme_Output_Port *)port;
  if (op->closed)
    return;

  if (op->close_fun)
    op->close_fun(op);
  op->closed = 1;
}

/* file input ports */

static int
file_char_ready (Scheme_Input_Port *port)
{
  FILE *fp;
  Scheme_Input_File *fip;

  fip = (Scheme_Input_File *)port->port_data;
  fp = fip->f;

#ifdef HAS_STANDARD_IOB  
  if (fp->_cnt)
    return 1;
#endif
#ifdef HAS_GNU_IOB
  if (fp->_egptr - fp->_gptr)
    return 1;
#endif
#ifdef HAS_LINUX_IOB
  if (fp->_IO_read_end - fp->_IO_read_ptr)
    return 1;
#endif
#ifdef HAS_BSD_IOB
  if (fp->_r >= 0) /* Submitted by Andreas Gustafsso, gson@araneus.pp.fi */
    return 1;
#endif

  if (feof(fp) || ferror(fp))
    return 1;

#ifdef FILES_HAVE_FDS
  {
    int fd;
    fd_set readfds;
    struct timeval time = {0, 0};

    fd = fileno(fp);
    
    FD_ZERO(&readfds);
    FD_SET(fd, &readfds);
    
    return select(fd + 1, &readfds, NULL, NULL, &time);
  }
#else
  return 1;
#endif
}

static int 
file_getc (Scheme_Input_Port *port)
{
  FILE *fp;
  Scheme_Input_File *fip;

  if ((Scheme_Object *)port == scheme_stdin_port) {
    /* Flush current output ports: */
    Scheme_Output_Port *op;
    
    op = (Scheme_Output_Port *)scheme_stdout_port;
    if (SAME_OBJ(op->sub_type, scheme_file_output_port_type))
      fflush((FILE *)op->port_data);
    
    op = (Scheme_Output_Port *)scheme_stderr_port;
    if (SAME_OBJ(op->sub_type, scheme_file_output_port_type))
      fflush((FILE *)op->port_data);
  }

  fip = (Scheme_Input_File *)port->port_data;
  fp = fip->f;

  if (!file_char_ready(port)) {
#ifdef FILES_HAVE_FDS
    scheme_current_process->block_descriptor = PORT_BLOCKED;
    scheme_current_process->blocker = (Scheme_Object *)port;
#endif
    do
      scheme_process_block(1);
    while (!file_char_ready(port));
#ifdef FILES_HAVE_FDS
    scheme_current_process->block_descriptor = NOT_BLOCKED;
    scheme_current_process->blocker = NULL;
    scheme_current_process->ran_some = 1;
#endif
  }

  return getc(fp);
}

static void
file_close_input (Scheme_Input_Port *port)
{
  Scheme_Input_File *fip;

  fip = (Scheme_Input_File *)port->port_data;
  fclose (fip->f);
}

static void
file_need_wakeup (Scheme_Input_Port *port, void *fds)
{
#ifdef FILES_HAVE_FDS
  Scheme_Input_File *fip;
  int n;

  fip = (Scheme_Input_File *)port->port_data;

  n = fileno(fip->f);
  FD_SET(n, (fd_set *)fds);
#endif
}

Scheme_Object *
scheme_make_named_file_input_port (FILE *fp, const char *filename)
{
  Scheme_Input_Port *ip;
  Scheme_Input_File *fip;

  fip = (Scheme_Input_File *)scheme_malloc(sizeof(Scheme_Input_File));

  fip->f = fp;

  ip =  scheme_make_input_port (scheme_file_input_port_type,
				fip,
				file_getc,
				file_char_ready,
				file_close_input,
				file_need_wakeup);

  ip->name = scheme_strdup(filename);

  return (Scheme_Object *)ip;
}

Scheme_Object *
scheme_make_file_input_port (FILE *fp)
{
  return scheme_make_named_file_input_port(fp, "FILE");
}

/* string input ports */

static int 
string_getc (Scheme_Input_Port *port)
{
  Scheme_Indexed_String *is;

  is = (Scheme_Indexed_String *) port->port_data;
  if (is->index >= is->size)
    {
      return (EOF);
    }
  else
    {
      return (is->string[is->index++]);
    }
}

static int
string_char_ready (Scheme_Input_Port *port)
{
  Scheme_Indexed_String *is;

  is = (Scheme_Indexed_String *) port->port_data;
  return (is->index < is->size);
}

static void
string_close_in (Scheme_Input_Port *port)
{
  return;
}

static Scheme_Indexed_String *
make_indexed_string (const char *str)
{
  Scheme_Indexed_String *is;

  is = (Scheme_Indexed_String *)scheme_malloc(sizeof(Scheme_Indexed_String));
  if (str) {
    is->string = scheme_strdup (str);
    is->size = strlen (str);
  } else {
    is->size = 100;
    is->string = (char *)scheme_malloc_atomic(is->size + 1);
  }
  is->index = 0;
  return (is);
}

Scheme_Object *
scheme_make_string_input_port(const char *str)
{
  Scheme_Input_Port *ip;

  ip = scheme_make_input_port (scheme_string_input_port_type,
			       make_indexed_string (str),
			       string_getc,
			       string_char_ready,
			       string_close_in,
			       NULL);

  ip->name = "STRING";

  return (Scheme_Object *)ip;
}

/* string output ports */

static void
string_write_string(char *str, long len, Scheme_Output_Port *port)
{
  Scheme_Indexed_String *is;

  is = (Scheme_Indexed_String *) port->port_data;

  if (is->index + len >= is->size) {
    char *old;

    old = is->string;

    if (len > is->size)
      is->size += 2 * len;
    else
      is->size *= 2;

    is->string = (char *)scheme_malloc_atomic(is->size + 1);
    memcpy(old, is->string, is->index);
  }
  
  memcpy(is->string + is->index, str, len);
  is->index += len;
}

static void
string_close_out (Scheme_Output_Port *port)
{
  return;
}

Scheme_Object *
scheme_make_string_output_port (void)
{
  Scheme_Output_Port *op;

  op = scheme_make_output_port (scheme_string_output_port_type,
				make_indexed_string(NULL),
				string_write_string,
				string_close_out);

  return (Scheme_Object *)op;
}

char *
scheme_get_string_output(Scheme_Object *port)
{
  Scheme_Output_Port *op;
  Scheme_Indexed_String *is;
  char *v;

  if (!SCHEME_OUTPORTP(port))
    return NULL;

  op = (Scheme_Output_Port *)port;
  if (op->sub_type != scheme_string_output_port_type)
    return NULL;

  is = (Scheme_Indexed_String *)op->port_data;

  v = scheme_malloc_atomic(is->size + 1);
  memcpy(v, is->string, is->size);
  v[is->size] = 0;

  return v;
}

/* user input ports */

static int 
user_getc (Scheme_Input_Port *port)
{
  Scheme_Object *fun, *val;

  fun = ((Scheme_Object **) port->port_data)[0];
  val = scheme_apply(fun, 0, NULL);
  if (SCHEME_EOFP(val))
    return EOF;
  else {
    SCHEME_ASSERT(SCHEME_CHARP(val), 
		  "port: user getc returned a non-character");
    return SCHEME_CHAR_VAL(val);
  }
}

static int
user_char_ready (Scheme_Input_Port *port)
{
  Scheme_Object *fun, *val;

  fun = ((Scheme_Object **) port->port_data)[1];
  val = scheme_apply(fun, 0, NULL);
  return SAME_OBJ(val, scheme_true);
}

static void
user_close_input (Scheme_Input_Port *port)
{
  Scheme_Object *fun;

  fun = ((Scheme_Object **) port->port_data)[2];
  scheme_apply(fun, 0, NULL);
}

/* file output ports */

static void
file_write_string (char *str, long len, Scheme_Output_Port *port)
{
  FILE *fp = (FILE *) port->port_data;

  fwrite(str, len, 1, fp);

  while (len--) {
    if (*(str++) == '\n') {
      fflush(fp);
      break;
    }
  }
}

static void
file_close_output (Scheme_Output_Port *port)
{
  FILE *fp = (FILE *) port->port_data;
  fclose (fp);
}

Scheme_Object *
scheme_make_file_output_port (FILE *fp)
{
  return (Scheme_Object *)scheme_make_output_port (scheme_file_output_port_type,
						   fp,
						   file_write_string,
						   file_close_output);
}

/* user output ports */

static void
user_write(char *str, long len, Scheme_Output_Port *port)
{
  Scheme_Object *fun, *p[1];
  
  p[0] = scheme_make_sized_string(str, len, 1);

  fun = ((Scheme_Object **) port->port_data)[0];
  scheme_apply(fun, 1, p);
}

static void
user_close_output (Scheme_Output_Port *port)
{
  Scheme_Object *fun;

  fun = ((Scheme_Object **) port->port_data)[1];
  scheme_apply(fun, 0, NULL);
}

/***********/

int scheme_file_exists(char *filename)
{
#ifdef NO_STAT
  FILE *fp;

  fp = fopen(filename, "r");
  if (fp) {
    fclose(fp);
    return 1;
  } else
    return 0;
#else
  struct stat buf;

  return !stat(filename, &buf);
#endif
}

static Scheme_Object *file_exists(int argc, Scheme_Object **argv)
{
  if (argc != 1)
    scheme_wrong_count("file-exists?", 1, 1, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("file-exists?", "string", 0, argc, argv);

  return scheme_file_exists(SCHEME_STR_VAL(argv[0])) ? 
    scheme_true : scheme_false;
}

static Scheme_Object *delete_file(int argc, Scheme_Object **argv)
{
  if (argc != 1)
    scheme_wrong_count("delete-file", 1, 1, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("delete-file", "string", 0, argc, argv);

  unlink(SCHEME_STR_VAL(argv[0]));

  return scheme_void;
}

/***********/

static Scheme_Object *
input_port_p (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("input-port?", 1, 1, argc, argv);

  return (SCHEME_INPORTP(argv[0]) ? scheme_true : scheme_false);
}

static Scheme_Object *
output_port_p (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
    scheme_wrong_count("output-port?", 1, 1, argc, argv);

  return (SCHEME_OUTPORTP(argv[0]) ? scheme_true : scheme_false);
}

static Scheme_Object *
current_input_port (int argc, Scheme_Object *argv[])
{
  if (argc && argc != 1)
    scheme_wrong_count("current-input-port", 0, 1, argc, argv);

  if (argc) {
    if (!SCHEME_INPORTP(argv[0]))
      scheme_wrong_type("current-input-port", "input port", 0, argc, argv);

    scheme_stdin_port = argv[0];
  }
  return scheme_stdin_port;
}

static Scheme_Object *
current_output_port (int argc, Scheme_Object *argv[])
{
  if (argc && argc != 1)
    scheme_wrong_count("current-output-port", 0, 1, argc, argv);

  if (argc) {
    if (!SCHEME_OUTPORTP(argv[0]))
      scheme_wrong_type("current-output-port", "output port", 0, argc, argv);

    scheme_stdout_port = argv[0];
  }
  return scheme_stdout_port;
}

static Scheme_Object *
current_error_port (int argc, Scheme_Object *argv[])
{
  if (argc && argc != 1)
    scheme_wrong_count("current-error-port", 0, 1, argc, argv);

  if (argc) {
    if (!SCHEME_OUTPORTP(argv[0]))
      scheme_wrong_type("current-error-port", "output port", 0, argc, argv);

    scheme_stderr_port = argv[0];
  }
  return scheme_stderr_port;
}

static Scheme_Object *
process_input_port (int argc, Scheme_Object *argv[])
{
  if (argc)
    scheme_wrong_count("process-input-port", 0, 0, argc, argv);
 
  return scheme_current_process->in_port;
}

static Scheme_Object *
process_output_port (int argc, Scheme_Object *argv[])
{
  if (argc)
    scheme_wrong_count("process-output-port", 0, 0, argc, argv);
 
  return scheme_current_process->out_port;
}

static Scheme_Object *
process_error_port (int argc, Scheme_Object *argv[])
{
  if (argc)
    scheme_wrong_count("process-error-port", 0, 0, argc, argv);
 
  return scheme_current_process->err_port;
}

static Scheme_Object *
make_input_port (int argc, Scheme_Object *argv[])
{
  Scheme_Input_Port *ip;
  Scheme_Object **copy;

  if (argc != 3)
     scheme_wrong_count("make-input-port", 3, 3, argc, argv);
  if (!SCHEME_PROCP(argv[0]))
    scheme_wrong_type("make-input-port", "procedure", 0, argc, argv);
  if (!SCHEME_PROCP(argv[1]))
    scheme_wrong_type("make-input-port", "procedure", 1, argc, argv);
  if (!SCHEME_PROCP(argv[2]))
    scheme_wrong_type("make-input-port", "procedure", 2, argc, argv);
  
  copy = (Scheme_Object **)scheme_malloc_stubborn(3 * sizeof(Scheme_Object *));
  memcpy(copy, argv, 3 * sizeof(Scheme_Object *));
  scheme_end_stubborn_change((void *)copy);

  ip = scheme_make_input_port (scheme_user_input_port_type,
			       copy,
			       user_getc,
			       user_char_ready,
			       user_close_input,
			       NULL);

  ip->name = "USERPORT";

  return (Scheme_Object *)ip;
}

static Scheme_Object *
make_output_port (int argc, Scheme_Object *argv[])
{
  Scheme_Output_Port *op;
  Scheme_Object **copy;

  if (argc != 2)
     scheme_wrong_count("make-output-port", 2, 2, argc, argv);
  if (!SCHEME_PROCP(argv[0]))
    scheme_wrong_type("make-output-port", "procedure", 0, argc, argv);
  if (!SCHEME_PROCP(argv[1]))
    scheme_wrong_type("make-output-port", "procedure", 1, argc, argv);

  copy = (Scheme_Object **)scheme_malloc_stubborn(2 * sizeof(Scheme_Object *));
  memcpy(copy, argv, 2 * sizeof(Scheme_Object *));
  scheme_end_stubborn_change((void *)copy);

  op = scheme_make_output_port(scheme_user_output_port_type,
			       copy,
			       user_write,
			       user_close_output);

  return (Scheme_Object *)op;
}

static Scheme_Object *
open_input_file (int argc, Scheme_Object *argv[])
{
  FILE *fp;
  char *mode = "rb";

  if (argc != 1 && argc != 2)
     scheme_wrong_count("open-input-file", 1, 2, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("open-input-file", "string", 0, argc, argv);
  if (argc > 1) {
    if (!SCHEME_SYMBOLP(argv[1]))
      scheme_wrong_type("open-input-file", "symbol", 1, argc, argv);
    
    if (SAME_OBJ(argv[1], text_symbol))
      mode = "rt";
    else if (SAME_OBJ(argv[1], binary_symbol)) {
      /* This is the default */
    } else
      scheme_signal_error("open-input-file: bad mode '%s",
			  SCHEME_SYM_VAL(argv[1]));
  }
  
  fp = fopen (SCHEME_STR_VAL(argv[0]), mode);
  if (!fp)
    scheme_signal_error ("Cannot open input file %s", SCHEME_STR_VAL(argv[0]));
  return scheme_make_named_file_input_port(fp, SCHEME_STR_VAL(argv[0]));
}

static Scheme_Object *
open_input_string (int argc, Scheme_Object *argv[])
{
  char *str;

  if (argc != 1)
     scheme_wrong_count("open-input-string", 1, 1, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("open-input-string", "string", 0, argc, argv);

  str = SCHEME_STR_VAL (argv[0]);
  return (scheme_make_string_input_port (str));
}

static Scheme_Object *
open_output_file (int argc, Scheme_Object *argv[])
{
  FILE *fp;
  int e_set = 0, m_set = 0, i;
  int existsok = 0;
  char *mode = "wb", *filename;
  
  if (argc < 1 && argc > 3)
     scheme_wrong_count("open-output-file", 1, 3, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("open-output-file", "string", 0, argc, argv);
  
  for (i = 1; argc > i; i++) {
    if (!SCHEME_SYMBOLP(argv[i]))
      scheme_wrong_type("open-output-file", "symbol", i, argc, argv);
  
    if (SAME_OBJ(argv[i], append_symbol)) {
      mode = "ab";
      existsok = -1;
      e_set++;
    } else if (SAME_OBJ(argv[i], replace_symbol)) {
      existsok = 1;
      e_set++;
    } else if (SAME_OBJ(argv[i], truncate_symbol)) {
      existsok = -1;
      e_set++;
    } else if (SAME_OBJ(argv[i], error_symbol)) {
      /* This is the default */
      e_set++;
    } else if (SAME_OBJ(argv[i], text_symbol)) {
      mode = "wt";
      m_set++;
    } else if (SAME_OBJ(argv[i], binary_symbol)) {
      /* This is the default */
      m_set++;
    } else
      scheme_signal_error("open-output-file: bad mode '%s",
			  SCHEME_SYM_VAL(argv[i]));

    if (m_set > 1 || e_set > 1)
      scheme_signal_error("open-output-file: conflicting or redundant "
			  "file modes given");
  }

  filename = SCHEME_STR_VAL(argv[0]);

  if (existsok >= 0) {
    if (scheme_file_exists(filename)) {
      if (!existsok)
	scheme_signal_error("open-output-file: file \"%s\" exists", filename);
      if (unlink(filename))
	scheme_signal_error("open-output-file: error deleting \"%s\"", 
			    filename);
    }
  }

  fp = fopen (filename, mode);
  if (!fp)
    scheme_signal_error("open-output-file: cannot open output file \"%s\"",
			filename);

  return scheme_make_file_output_port(fp);
}

static Scheme_Object *
open_output_string (int argc, Scheme_Object *argv[])
{
  if (argc != 0)
     scheme_wrong_count("open-output-string", 0, 0, argc, argv);

  return scheme_make_string_output_port();
}

static Scheme_Object *
get_output_string (int argc, Scheme_Object *argv[])
{
  Scheme_Output_Port *op;
  Scheme_Indexed_String *is;

  if (argc != 1)
     scheme_wrong_count("get-output-string", 1, 1, argc, argv);
  if (!SCHEME_OUTPORTP(argv[0]))
    scheme_wrong_type("get-output-string", "output port", 0, argc, argv);

  op = (Scheme_Output_Port *)argv[0];
  if (op->sub_type != scheme_string_output_port_type)
    scheme_signal_error("get-output-string: wrong type of port");

  is = (Scheme_Indexed_String *)op->port_data;
  is->string[is->index] = 0;

  return scheme_make_string(is->string);
}

static Scheme_Object *
close_input_port (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
     scheme_wrong_count("close-input-port", 1, 1, argc, argv);
  if (!SCHEME_INPORTP(argv[0]))
    scheme_wrong_type("close-input-port", "input port", 0, argc, argv);

  scheme_close_input_port (argv[0]);
  return (scheme_void);
}

static Scheme_Object *
close_output_port (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
     scheme_wrong_count("close-output-port", 1, 1, argc, argv);
  if (!SCHEME_OUTPORTP(argv[0]))
    scheme_wrong_type("close-output-port", "output port", 0, argc, argv);

  scheme_close_output_port (argv[0]);
  return (scheme_void);
}

static Scheme_Object *
read_f (int argc, Scheme_Object *argv[])
{
  Scheme_Object *port;

  if (argc && argc != 1)
     scheme_wrong_count("read", 0, 1, argc, argv);
  if (argc && !SCHEME_INPORTP(argv[0]))
    scheme_wrong_type("read", "input port", 0, argc, argv);

  port = (argc ? argv[0] : scheme_stdin_port);
  
  return scheme_read(port);
}

static Scheme_Object *
do_read_char (char *name, int argc, Scheme_Object *argv[], int peek)
{
  Scheme_Object *port;
  int ch;

  if (argc && argc != 1)
     scheme_wrong_count(name, 0, 1, argc, argv);
  if (argc && !SCHEME_INPORTP(argv[0]))
    scheme_wrong_type(name, "input port", 0, argc, argv);

  port = (argc ? argv[0] : scheme_stdin_port);
  
  ch = scheme_getc(port);

  if (ch == EOF)
    return (scheme_eof);
  else {
    if (peek)
      scheme_ungetc(ch, port);
    return scheme_make_char(ch);
  }
}
static Scheme_Object *
read_char (int argc, Scheme_Object *argv[])
{
  return do_read_char("read-char", argc, argv, 0);
}

static Scheme_Object *
peek_char (int argc, Scheme_Object *argv[])
{
  return do_read_char("peek-char", argc, argv, 1);
}

static Scheme_Object *
read_line (int argc, Scheme_Object *argv[])
{
  Scheme_Object *port;
  int ch;
  char *buf, *oldbuf, onstack[32];
  long size = 31, oldsize, i = 0;

  if (argc && argc != 1)
     scheme_wrong_count("read-line", 0, 1, argc, argv);
  if (argc && !SCHEME_INPORTP(argv[0]))
    scheme_wrong_type("read-line", "input port", 0, argc, argv);

  port = (argc ? argv[0] : scheme_stdin_port);
  
  buf = onstack;

  while (1) {
    ch = scheme_getc(port);
    if (ch == EOF) {
      if (!i)
	return scheme_eof;
      break;
    }
    if (ch == '\r' || ch == '\n')
      break;
    
    if (i >= size) {
      oldsize = size;
      oldbuf = buf;

      size *= 2;
      buf = (char *)scheme_malloc_atomic(size + 1);
      memcpy(buf, oldbuf, oldsize);
    }
    buf[i++] = ch;
  }
  buf[i] = '\0';

  return scheme_make_sized_string(buf, i, buf == onstack);
}

static Scheme_Object *
eof_object_p (int argc, Scheme_Object *argv[])
{
  if (argc != 1)
     scheme_wrong_count("eof-object?", 0, 1, argc, argv);

  return (SCHEME_EOFP(argv[0]) ? scheme_true : scheme_false);
}

static Scheme_Object *
char_ready_p (int argc, Scheme_Object *argv[])
{
  Scheme_Object *port;

  if (argc && argc != 1)
     scheme_wrong_count("char-ready?", 0, 1, argc, argv);
  if (argc && !SCHEME_INPORTP(argv[0]))
    scheme_wrong_type("char-ready?", "input port", 0, argc, argv);

  port = (argc ? argv[0] : scheme_stdin_port);
  
  return (scheme_char_ready(port) ? scheme_true : scheme_false);
}

static Scheme_Object *
display_write(char *name, int argc, Scheme_Object *argv[], int escape)
{
  Scheme_Object *port;

  if (argc != 1 && argc != 2)
     scheme_wrong_count(name, 1, 2, argc, argv);
  if (argc > 1 && !SCHEME_OUTPORTP(argv[1]))
    scheme_wrong_type(name, "output port", 1, argc, argv);

  port = (argc > 1 ? argv[1] : scheme_stdout_port);
  
  if (escape)
    scheme_display(argv[0], port);
  else
    scheme_write(argv[0], port);

  return scheme_void;
}

static Scheme_Object *
sch_write (int argc, Scheme_Object *argv[])
{
  return display_write("write", argc, argv, 0);
}

static Scheme_Object *
display (int argc, Scheme_Object *argv[])
{
  return display_write("display", argc, argv, 1);
}

static Scheme_Object *
newline (int argc, Scheme_Object *argv[])
{
  Scheme_Object *port;

  if (argc != 0 && argc != 1)
     scheme_wrong_count("newline", 0, 1, argc, argv);
  if (argc && !SCHEME_OUTPORTP(argv[0]))
    scheme_wrong_type("newline", "output port", 0, argc, argv);

  port = (argc ? argv[0] : scheme_stdout_port);
  
  scheme_write_string("\n", 1, port);

  return scheme_void;
}

static Scheme_Object *
write_char (int argc, Scheme_Object *argv[])
{
  if (argc && !SCHEME_CHARP(argv[0]))
    scheme_wrong_type("write-char", "character", 0, argc, argv);
  
  return display_write("write-char", argc, argv, 1);
}

static Scheme_Object *
load (int argc, Scheme_Object *argv[])
{
  jmp_buf savebuf;
  Scheme_Object *obj, *port;
  char *filename;
  int ch, crc;
  FILE *fp;

  if (argc != 1)
     scheme_wrong_count("load", 1, 1, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("load", "string", 0, argc, argv);

  filename = SCHEME_STR_VAL (argv[0]);
  fp = fopen (filename, "rb");
  if (! fp)
    scheme_signal_error ("load: could not open file for input: %s", filename);

  memcpy(&savebuf, &scheme_error_buf, sizeof(jmp_buf));
  crc = scheme_can_read_compiled;
  if (setjmp(scheme_error_buf)) {
    fclose(fp);
    scheme_can_read_compiled = crc;
    scheme_longjmp(savebuf, 1);
  } else {
    port = scheme_make_named_file_input_port (fp, filename);
    /* Skip over #! at beginning of file */
    if ((ch = scheme_getc(port)) == '#') {
      if ((ch = scheme_getc(port)) == '!') {
	int oldch = 0;
      eol_loop:
	while ((ch = scheme_getc(port)) != '\n')
	  oldch = ch;
	if (oldch == '\\')
	  goto eol_loop;
      } else {
	scheme_ungetc(ch, port);
	scheme_ungetc('#', port);
      }
    } else
      scheme_ungetc(ch, port);

    scheme_can_read_compiled = 1;
    while ((obj = scheme_read (port)) && !SCHEME_EOFP(obj))
      scheme_evaluator(obj);
    scheme_can_read_compiled = crc;

    fclose (fp);
    memcpy(&scheme_error_buf, &savebuf, sizeof(jmp_buf));
  }

  return scheme_void;
}

int scheme_load(const char *file)
{
  Scheme_Object *p[1];
  jmp_buf savebuf;
  int success;

  p[0] = scheme_make_string(file);
  memcpy(&savebuf, &scheme_error_buf, sizeof(jmp_buf));
  if (setjmp(scheme_error_buf)) {
    success = 0;
  } else {
    scheme_apply(load_prim, 1, p);
    success = 1;
  }
  memcpy(&scheme_error_buf, &savebuf, sizeof(jmp_buf));

  return success;
}

static Scheme_Object *
flush_output (int argc, Scheme_Object *argv[])
{
  Scheme_Output_Port *op;

  if (argc > 1)
     scheme_wrong_count("flush-output", 0, 1, argc, argv);
  if (argc && !SCHEME_OUTPORTP(argv[0]))
    scheme_wrong_type("flush-output", "output port", 0, argc, argv);

  op = (Scheme_Output_Port *)(argc ? argv[0] : scheme_stdout_port);

  if (SAME_OBJ(op->sub_type, scheme_file_output_port_type))
    fflush ((FILE *)op->port_data);

  return (scheme_void);
}

static Scheme_Object *
file_position(int argc, Scheme_Object *argv[])
{
  FILE *f;

  if (argc != 1 && argc != 2)
     scheme_wrong_count("file-position", 1, 2, argc, argv);
  if (!SCHEME_OUTPORTP(argv[0]) && !SCHEME_INPORTP(argv[0]))
    scheme_wrong_type("file-position", "port", 0, argc, argv);
  if (argc == 2 && !SCHEME_INTP (argv[1]))
    scheme_wrong_type("file-position", "integer", 1, argc, argv);

  f = NULL;
  if (SCHEME_OUTPORTP(argv[0])) {
    Scheme_Output_Port *op;

    op = (Scheme_Output_Port *)argv[0];
    if (SAME_OBJ(op->sub_type, scheme_file_output_port_type))
      f = (FILE *)op->port_data;
  } else if (SCHEME_INPORTP(argv[0])) {
    Scheme_Input_Port *ip;
  
    ip = (Scheme_Input_Port *)argv[0];
    if (SAME_OBJ(ip->sub_type, scheme_file_input_port_type))
      f = ((Scheme_Input_File *)ip->port_data)->f;
    else if (!argc)
      return scheme_make_integer(scheme_tell(argv[0]));
  }

  if (!f)
    scheme_signal_error("file-position: must specify a file port");

  if (argc > 1) {
    fseek(f, SCHEME_INT_VAL(argv[1]), 0);
    return scheme_void;
  } else {
    int p;
    p = ftell(f);
    return scheme_make_integer(p);
  }
}

void scheme_flush_output(Scheme_Object *port)
{
  (void)flush_output(1, &port);
}


typedef struct {
  char *buf;
  long buflen;
  int bufstart, bufend;
  int eof;
} Scheme_Pipe;

int pipe_getc(Scheme_Input_Port *p)
{
  Scheme_Pipe *pipe;
  int c;

  pipe = (Scheme_Pipe *)(p->port_data);
  
  scheme_current_process->block_descriptor = PIPE_BLOCKED;
  while (pipe->bufstart == pipe->bufend && !pipe->eof)
    scheme_process_block(1);
  scheme_current_process->block_descriptor = NOT_BLOCKED;
  scheme_current_process->ran_some = 1;

  if (pipe->bufstart == pipe->bufend)
    return EOF;

  c = pipe->buf[pipe->bufstart];

  pipe->bufstart = (pipe->bufstart + 1) % pipe->buflen;

  return c;
}

void pipe_write(char *str, long len, Scheme_Output_Port *p)
{
  Scheme_Pipe *pipe;
  long avail, firstpos, firstn, secondpos, secondn, endpos;

  pipe = (Scheme_Pipe *)(p->port_data);
  
  if (pipe->eof)
    return;

  if (pipe->bufstart <= pipe->bufend) {
    firstn = pipe->buflen - pipe->bufend;
    avail = firstn + pipe->bufstart - 1;
  } else {
    firstn = avail = pipe->bufstart - pipe->bufend - 1;
  }
  firstpos = pipe->bufend;

  if (avail < len) {
    char *old;
    int newlen;

    old = pipe->buf;
    newlen = pipe->buflen + 2 * len;

    pipe->buf = scheme_malloc_atomic(newlen);

    if (pipe->bufstart <= pipe->bufend) {
      memcpy(pipe->buf, old + pipe->bufstart, pipe->bufend - pipe->bufstart);
      pipe->bufend -= pipe->bufstart;
      pipe->bufstart = 0;
    } else {
      int slen;
      slen = pipe->buflen - pipe->bufstart;
      memcpy(pipe->buf, old + pipe->bufstart, slen);
      memcpy(pipe->buf + slen, old, pipe->bufend);
      pipe->bufstart = 0;
      pipe->bufend += slen;
    }

    pipe->buflen = newlen;

    firstpos = pipe->bufend;
    firstn = len;
    endpos = firstpos + firstn;

    secondn = secondpos = 0;
  } else {
    if (firstn > len) {
      firstn = len;
      endpos = firstpos + len;
      secondn = secondpos = 0;
    } else {
      secondn = len - firstn;
      secondpos = 0;
      endpos = secondn;
    }
  }

  if (firstn)
    memcpy(pipe->buf + firstpos, str, firstn);
  if (secondn)
    memcpy(pipe->buf + secondpos, str + firstn, secondn);

  pipe->bufend = endpos;
}

int pipe_char_ready(Scheme_Input_Port *p)
{
  Scheme_Pipe *pipe;

  pipe = (Scheme_Pipe *)(p->port_data);
  
  return (pipe->bufstart != pipe->bufend || pipe->eof);
}

void pipe_in_close(Scheme_Input_Port *p)
{
  Scheme_Pipe *pipe;

  pipe = (Scheme_Pipe *)(p->port_data);
  
  pipe->eof = 1;
}

void pipe_out_close(Scheme_Output_Port *p)
{
  Scheme_Pipe *pipe;

  pipe = (Scheme_Pipe *)(p->port_data);
  
  pipe->eof = 1;
}

void scheme_pipe(Scheme_Object **read, Scheme_Object **write)
{
  Scheme_Pipe *pipe;
  Scheme_Input_Port *readp;
  Scheme_Output_Port *writep;

  pipe = (Scheme_Pipe *)scheme_malloc(sizeof(Scheme_Pipe));
  pipe->buflen = 100;
  pipe->buf = (char *)scheme_malloc_atomic(pipe->buflen);
  pipe->bufstart = pipe->bufend = 0;
  pipe->eof = 0;

  readp = scheme_make_input_port (pipe_read_port_type,
				  (void *)pipe,
				  pipe_getc,
				  pipe_char_ready,
				  pipe_in_close,
				  NULL);

  readp->name = "PIPE";

  writep = scheme_make_output_port (pipe_write_port_type,
				    (void *)pipe,
				    pipe_write,
				    pipe_out_close);

  *read = (Scheme_Object *)readp;
  *write = (Scheme_Object *)writep;
}

static Scheme_Object *sch_pipe(int argc, Scheme_Object **args)
{
  Scheme_Object *read, *write;

  if (argc)
    scheme_wrong_count("pipe", 0, 0, argc, args);

  scheme_pipe(&read, &write);

  return scheme_make_pair(read, scheme_make_pair(write, scheme_null));
}

#ifdef DIR_FUNCTION
static Scheme_Object *current_directory(int argc, Scheme_Object **args)
{
  if (argc && argc != 1)
    scheme_wrong_count("current-directory", 0, 1, argc, args);

  if (argc) {
    if (!SCHEME_STRINGP(args[0]))
      scheme_wrong_type("current-directory", "string", 0, argc, args);

    chdir(SCHEME_STR_VAL(args[0]));

    return scheme_void;
  } else {
    char dir[1000];

    if (getcwd(dir, 999) == NULL) {
      dir[0] = '.';
      dir[1] = '\0';
    }

    return scheme_make_string(dir);
  }
}
#endif

#ifdef PROCESS_FUNCTION

#include <signal.h>

static Scheme_Object *process(int c, Scheme_Object *args[], 
			      char *name, int shell, int synchonous)
{
  char *command;
  int to_subprocess[2], from_subprocess[2], i, pid;
  char **argv;
  Scheme_Object *in, *out, *subpid;

  if (shell) {
    if (c != 1)
      scheme_wrong_count(name, 1, 1, c, args);
    if (!SCHEME_STRINGP(args[0]))
      scheme_wrong_type(name, "string", 0, c, args);
    argv = NULL;
  } else  {
    if (c < 1)
      scheme_wrong_count(name, 1, -1, c, args);

    argv = (char **)scheme_malloc_atomic((c + 1) * sizeof(char *));
    for (i = 0; i < c; i++) {
      if (!SCHEME_STRINGP(args[i]))
	scheme_wrong_type(name, "string", i, c, args);
      argv[i] = SCHEME_STR_VAL(args[i]);
    }
    argv[c] = NULL;
  }

  command = SCHEME_STR_VAL(args[0]);

  if (!synchonous) {
    pipe (to_subprocess);
    pipe (from_subprocess);
  }

  pid = fork ();

  switch (pid)
    {
    case -1:
      return scheme_false;

    case 0:			/* child */

      if (!synchonous) {
	/* Copy pipe descriptors to stdin and stdout */
	dup2 (to_subprocess[0], 0);
	dup2 (from_subprocess[1], 1);
	
	/* Close unwanted descriptors */
	
	close (to_subprocess[0]);
	close (to_subprocess[1]);
	close (from_subprocess[0]);
	close (from_subprocess[1]);
      }

      /* Exec new process */

      if (shell) {
	int v;

	if ((v = system(command)))
	  scheme_console_printf("shell startup failed");

	_exit(v);
      } else {
	int err;

	err = execv(command, argv);

	/* If we get here it failed; give up */

	scheme_console_printf("exec of %s failed (%ld)", command, err);
	_exit(1);
      }

    default:		/* parent */

      break;
    }

  /* Close unneeded descriptors */

  if (!synchonous) {
    close (to_subprocess[0]);
    close (from_subprocess[1]);

    in = scheme_make_file_input_port(fdopen(from_subprocess[0], "r"));
    out = scheme_make_file_output_port(fdopen(to_subprocess[1], "w"));
    subpid = scheme_make_integer(pid);

    return scheme_make_pair(in,
			    scheme_make_pair(out,
					     scheme_make_pair(subpid, 
							      scheme_null)));
  } else {
    int status;

#if !defined(__osf__) && !defined(SVR4) && !defined(sun) && !defined(__sgi) && !defined(__hpux) || defined(__SUNPRO_CC)
    while (wait ((union wait *) &status) != pid);
#else
    while (wait (&status) != pid);
#endif

    return status ? scheme_false : scheme_true;
  }
}

Scheme_Object *sch_process_star(int c, Scheme_Object *args[])
{
  return process(c, args, "process*", 0, 0);
}

Scheme_Object *sch_system_star(int c, Scheme_Object *args[])
{
  return process(c, args, "system*", 0, 1);
}

Scheme_Object *sch_process(int c, Scheme_Object *args[])
{
  return process(c, args, "process", 1, 0);
}

Scheme_Object *sch_system(int c, Scheme_Object *args[])
{
  return process(c, args, "system", 1, 1);
}

#endif

static Scheme_Object *read_accept_compiled(int argc, Scheme_Object *argv[])
{
  if (argc && argc != 1)
     scheme_wrong_count("read-accept-compiled", 0, 1, argc, argv);

  if (argc) {
    scheme_can_read_compiled = NOT_SAME_OBJ(argv[0], scheme_false);
    return scheme_void;
  } else {
    return scheme_can_read_compiled ? scheme_true : scheme_false;
  }
}
