/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.
*/

#include "schpriv.h"
#include <string.h>
#include <memory.h>

Scheme_Type_Reader *scheme_type_readers;
Scheme_Type_Writer *scheme_type_writers;

static char **type_names;
static Scheme_Type maxtype, allocmax;

void
scheme_init_type (Scheme_Env *env)
{
  long n;

  if (!scheme_starting_up)
    return;

  maxtype = _scheme_last_type_;
  allocmax = maxtype + 10;

  type_names = (char **)scheme_malloc(allocmax * sizeof(char *));
  scheme_type_readers = 
    (Scheme_Type_Reader *)scheme_malloc_atomic(n = allocmax
					       * sizeof(Scheme_Type_Reader));
  memset((char *)scheme_type_readers, 0, n);

  scheme_type_writers = 
    (Scheme_Type_Writer *)scheme_malloc_atomic(n = allocmax
					       * sizeof(Scheme_Type_Writer));
  memset((char *)scheme_type_writers, 0, n);

#define set_name(t, n) type_names[t] = n

  set_name(scheme_true_type, "<true>");
  set_name(scheme_false_type, "<false>");
  set_name(scheme_char_type, "<char>");
  set_name(scheme_debug_handle_type, "<debug-handle>");
  set_name(scheme_envunbox_type, "<env-auto-unbox>");
  set_name(scheme_static_distance_type, "<static-distance>");
  set_name(scheme_static_distance_unbox_type, "<static-distance>");
  set_name(scheme_variable_type, "<variable>");
  set_name(scheme_application_type, "<application>");
  set_name(scheme_closure_compilation_type, "<closure-compilation>");
  set_name(scheme_syntax_compilation_type, "<syntax-compilation>");
  set_name(scheme_branch_compilation_type, "<branch-compilation>");
  set_name(scheme_sequence_compilation_type, "<sequence-compilation>");
  set_name(scheme_let_value_compilation_type, "<let-value-compilation>");
  set_name(scheme_let_void_compilation_type, "<let-void-compilation>");
  set_name(scheme_eval_waiting_type, "<eval-waiting>");
  set_name(scheme_void_type, "<void>");
  set_name(scheme_prim_type, "<primitive>");
  set_name(scheme_closed_prim_type, "<closed-primitive>");
  set_name(scheme_closure_type, "<closure-form>");
  set_name(scheme_compiled_closure_type, "<procedure>");
  set_name(scheme_cont_type, "<continuation>");
  set_name(scheme_tail_call_waiting_type, "<tail call waiting>");
  set_name(scheme_null_type, "<empty-list>");
  set_name(scheme_pair_type, "<pair>");
  set_name(scheme_box_type, "<box>");
  set_name(scheme_integer_type, "<integer>");
  set_name(scheme_double_type, "<double>");
  set_name(scheme_object_type, "<object>");
  set_name(scheme_class_type, "<class>");
  set_name(scheme_class_data_type, "<class-data>");
  set_name(scheme_object_frame_type, "<object-frame>");
  set_name(scheme_generic_type, "<unknown-external-type>");
  set_name(scheme_undefined_type, "<undefined>");
  set_name(scheme_inherit_filler_type, "<inherited-undefined>");
  set_name(scheme_eof_type, "<eof>");
  set_name(scheme_input_port_type, "<input-port>");
  set_name(scheme_output_port_type, "<output-port>");
  set_name(scheme_process_type, "<process>");
  set_name(scheme_promise_type, "<promise>");
  set_name(scheme_string_type, "<string>");
  set_name(scheme_struct_proc_type, "<struct-procedure>");
  set_name(scheme_struct_info_type, "<structure-info>");
  set_name(scheme_structure_type, "<structure>");
  set_name(scheme_symbol_type, "<symbol>");
  set_name(scheme_compiled_syntax_type, "<syntax>");
  set_name(scheme_macro_type, "<macro>");
  set_name(scheme_let_data_type, "<let-data>");
  set_name(scheme_compiled_unquote_splicing_type, "<compiled-unquote-splicing>");
  set_name(scheme_compiled_unquote_type, "<compiled-unquote>");
  set_name(scheme_compiled_quoted_vector_type, "<compiled-quoted-vector>");
  set_name(scheme_vector_type, "<vector>");
  set_name(scheme_type_symbol_type, "<type-symbol>");
  set_name(scheme_bignum_type, "<bignum>");
  set_name(scheme_escaping_cont_type, "<escape-continuation>");
  set_name(scheme_trigger_type, "<trigger>");
  set_name(scheme_hash_table_type, "<hash-table>");
  set_name(scheme_case_closure_type, "<procedure*>");
  set_name(scheme_generic_data_type, "<generic-data>");
  set_name(scheme_multiple_values_type, "<multiple-values>");
  set_name(scheme_delayed_ivar_type, "<delayed-ivar>");

  set_name(scheme_reserved_6_type, "<reserved6>");
  set_name(scheme_reserved_7_type, "<reserved7>");
  set_name(scheme_reserved_8_type, "<reserved8>");
  set_name(scheme_reserved_9_type, "<reserved9>");
  set_name(scheme_reserved_10_type, "<reserved10>");
}

Scheme_Type scheme_make_type(const char *name)
{
  if (maxtype == allocmax) {
    /* Expand arrays */
    void *naya;
    long n;
    
    allocmax += 20;

    naya = scheme_malloc(allocmax * sizeof(char *));
    memcpy(naya, type_names, maxtype * sizeof(char *));
    type_names = (char **)naya;

    naya = scheme_malloc_atomic(n = allocmax * sizeof(Scheme_Type_Reader));
    memset((char *)naya, 0, n);
    memcpy(naya, scheme_type_readers, maxtype * sizeof(Scheme_Type_Reader));
    scheme_type_readers = (Scheme_Type_Reader *)naya;

    naya = scheme_malloc_atomic(n = allocmax * sizeof(Scheme_Type_Reader));
    memset((char *)naya, 0, n);
    memcpy(naya, scheme_type_writers, maxtype * sizeof(Scheme_Type_Reader));
    scheme_type_writers = (Scheme_Type_Writer *)naya;
  }
  
  type_names[maxtype] = scheme_strdup(name);

  return maxtype++;
}

char *scheme_get_type_name(Scheme_Type t)
{
  if (t < 0 || t >= maxtype)
    return "<bad-type>";
  return type_names[t];
}

int scheme_find_type(Scheme_Object *ts)
{
  int i;
  char *str;

  str = SCHEME_TSYM_VAL(ts);

  for (i = maxtype; i--; ) {
    if (type_names[i]) {
      char *ts = type_names[i] + 1, *s = str;
      while (*s && *ts && (*s == *ts)) {
	s++;
	ts++;
      }
      if (!*s && *ts == '>')
	return i;
    }
  }

  scheme_signal_error("bad type name: %s", str);
  return -1;
}

void scheme_install_type_reader(Scheme_Type t, Scheme_Type_Reader f)
{
  if (t < 0 || t >= maxtype)
    return;

  scheme_type_readers[t] = f;
}

void scheme_install_type_writer(Scheme_Type t, Scheme_Type_Writer f)
{
  if (t < 0 || t >= maxtype)
    return;

  scheme_type_writers[t] = f;
}
