;;;; `test.scm' Test correctness of scheme implementations.
;;; Copyright (C) 1991, 1992, 1993, 1994 Aubrey Jaffer.

;;; MODIFIED for MzScheme - Matthew 8/95
;;;  Added a few more tests, like append!, reverse!, etc.
;;;  Added testing for inexact numbers
;;;  Added a lot of error testing
;;; Still need bignum testing

;;; This includes examples from
;;; William Clinger and Jonathan Rees, editors.
;;; Revised^4 Report on the Algorithmic Language Scheme
;;; and the IEEE specification.

;;; The input tests read this file expecting it to be named "test.scm".
;;; Files `tmp1', `tmp2' and `tmp3' will be created in the course of running
;;; these tests.  You may need to delete them in order to run
;;; "test.scm" more than once.

;;; There are three optional tests: (test-cont) tests multiple returns
;;; from a call-with-current-continuation; (test-sc4) tests
;;; procedures required by R4RS but not by IEEE; (test-inexact) tests
;;; inexact numbers.
;;; If you are testing a R3RS version which does not have `list?' do:
;;; (define list? #f)

;;; send corrections or additions to jaffer@ai.mit.edu or
;;; Aubrey Jaffer, 84 Pleasant St., Wakefield MA 01880, USA

(define cur-section '())(define errs '())
(define SECTION (lambda args
		  (display "SECTION") (write args) (newline)
		  (set! cur-section args) #t))
(define record-error (lambda (e) (set! errs (cons (list cur-section e) errs))))

(define test
  (lambda (expect fun . args)
    (write (cons fun args))
    (display "  ==> ")
    ((lambda (res)
      (write res)
      (newline)
      (cond ((not (equal? expect res))
	     (record-error (list res expect (cons fun args)))
	     (display " BUT EXPECTED ")
	     (write expect)
	     (newline)
	     #f)
	    (else #t)))
     (if (procedure? fun) (apply fun args) (car args)))))
(define error-test
  (lambda (expr)
    (write expr)
    (display "  ==> ")
    (let ([error?
	   (catch escape
	      (let ([old-handler (error-handler)]
		    [test-handler
		     (lambda (d s . rest)
		       (display "error: ")
		       (write s)
		       (newline)
		       (escape #t))])
		(dynamic-wind
		 (lambda () (error-handler test-handler))
		 (lambda ()
		   (let ([v (eval expr)])
		     (write v)
		     (display " BUT EXPECTED ERROR")
		     (newline)
		     #f))
		 (lambda () (error-handler old-handler)))))])
      (if (not error?)
	  (record-error (list 'Error expr))))))
(define (arity-test f min max)
  (let loop ([n 0][l '()])
    (unless (>= n min)
       (error-test `(,f ,@l))
       (loop (add1 n) (cons 1 l))))
  (if (> max 0)
      (error-test `(,f ,@(let loop ([n 0][l '(1)])
			   (if (= n max)
			       l
			       (loop (add1 n) (cons 1 l))))))))
(define (report-errs)
  (newline)
  (if (null? errs) (display "Passed all tests")
      (begin
	(display "errors were:")
	(newline)
	(display "(SECTION (got expected (call)))")
	(newline)
	(for-each (lambda (l) (write l) (newline))
		  errs)))
  (newline))

(SECTION 'STRUCTURE)
(define-struct a (b c))
(define-struct aa ())
(define ai (make-a 1 2))
(define aai (make-aa))
(test #t procedure? a?)
(test #t a? ai)
(test #f a? 1)
(test #f aa? ai)
(test 1 a-b ai)
(test 2 a-c ai)
(define ai2 (make-a 1 2))
(set-a-b! ai2 3)
(set-a-c! ai2 4)
(test 1 a-b ai)
(test 2 a-c ai)
(test 3 a-b ai2)
(test 4 a-c ai2)
(define-struct a (b c))
(test #f a? ai)
(arity-test make-a 2 2)
(error-test '(make-aa 1))
(arity-test a? 1 1)
(arity-test a-b 1 1)
(arity-test a-c 1 1)
(arity-test set-a-b! 2 2)
(arity-test set-a-c! 2 2)
(error-test '(a-b 5))
(error-test '(a-b ai))
(error-test '(set-a-b! ai 5))
(error-test '(set-a-c! ai 5))

(SECTION 2 1);; test that all symbol characters are supported.
'(+ - ... !.. $.+ %.- &.! *.: /:. :+. <-. =. >. ?. ~. _. ^.)

(SECTION 3 4)
(define disjoint-type-functions
  (list boolean? char? null? number? pair? procedure? string? symbol? vector?))
(define type-examples
  (list
   #t #f #\a '() 9739 '(test) record-error "test" "" 'test '#() '#(a b c) ))
(define i 1)
(for-each (lambda (x) (display (make-string i #\ ))
		  (set! i (+ 3 i))
		  (write x)
		  (newline))
	  disjoint-type-functions)
(define type-matrix
  (map (lambda (x)
	 (let ((t (map (lambda (f) (f x)) disjoint-type-functions)))
	   (write t)
	   (write x)
	   (newline)
	   t))
       type-examples))
(SECTION 4 1 2)
(test '(quote a) 'quote (quote 'a))
(test '(quote a) 'quote ''a)
(error-test '(quote))
(error-test '(quote 1 2))
(SECTION 4 1 3)
(test 12 (if #f + *) 3 4)
(error-test '(+ 3 . 4))
(SECTION 4 1 4)
(test 8 (lambda (x) (+ x x)) 4)
(define reverse-subtract
  (lambda (x y) (- y x)))
(test 3 reverse-subtract 7 10)
(define add4
  (let ((x 4))
    (lambda (y) (+ x y))))
(test 10 add4 6)
(test '(3 4 5 6) (lambda x x) 3 4 5 6)
(test '(5 6) (lambda (x y . z) z) 3 4 5 6)
(error-test '(lambda))
(error-test '(lambda x))
(error-test '(lambda . x))
(error-test '(lambda x . x))
(error-test '(lambda ((x)) x))
(error-test '(lambda 5 x))
(error-test '(lambda (5) x))
(error-test '(lambda (x (y)) x))
(error-test '(lambda (x . 5) x))
(SECTION 4 1 5)
(test 'yes 'if (if (> 3 2) 'yes 'no))
(test 'no 'if (if (> 2 3) 'yes 'no))
(test '1 'if (if (> 3 2) (- 3 2) (+ 3 2)))
(error-test '(if . #t))
(error-test '(if #t . 1))
(error-test '(if #t 1 . 2))
(error-test '(if #t))
(error-test '(if #t 1 2 3))
(SECTION 4 1 6)
(define x 2)
(test 3 'define (+ x 1))
(set! x 4)
(test 5 'set! (+ x 1))
(error-test '(set!))
(error-test '(set! x))
(error-test '(set! x 1 2))
(error-test '(set! 1 2))
(error-test '(set! (x) 1))
(error-test '(set! . x))
(error-test '(set! x . 1))
(SECTION 4 2 1)
(test 'greater 'cond (cond ((> 3 2) 'greater)
			   ((< 3 2) 'less)))
(test 'equal 'cond (cond ((> 3 3) 'greater)
			 ((< 3 3) 'less)
			 (else 'equal)))
(test 2 'cond (cond ((assv 'b '((a 1) (b 2))) => cadr)
		     (else #f)))
(test 'composite 'case (case (* 2 3)
			 ((2 3 5 7) 'prime)
			 ((1 4 6 8 9) 'composite)))
(test 'consonant 'case (case (car '(c d))
			 ((a e i o u) 'vowel)
			 ((w y) 'semivowel)
			 (else 'consonant)))
(test #t 'and (and (= 2 2) (> 2 1)))
(test #f 'and (and (= 2 2) (< 2 1)))
(test '(f g) 'and (and 1 2 'c '(f g)))
(test #t 'and (and))
(test #t 'or (or (= 2 2) (> 2 1)))
(test #t 'or (or (= 2 2) (< 2 1)))
(test #f 'or (or #f #f #f))
(test #f 'or (or))
(test '(b c) 'or (or (memq 'b '(a b c)) (+ 3 0)))
(error-test '(cond #t))
(error-test '(cond ()))
(error-test '(cond . #t))
(error-test '(cond (#t . 1)))
(error-test '(case))
(error-test '(case 0 #t))
(error-test '(case 0 ()))
(error-test '(and . 1))
(error-test '(and 1 . 2))
(error-test '(or . 1))
(error-test '(or 1 . 2))
(SECTION 4 2 2)
(test 6 'let (let ((x 2) (y 3)) (* x y)))
(test 35 'let (let ((x 2) (y 3)) (let ((x 7) (z (+ x y))) (* z x))))
(test 70 'let* (let ((x 2) (y 3)) (let* ((x 7) (z (+ x y))) (* z x))))
(test #t 'letrec (letrec ((even?
			   (lambda (n) (if (zero? n) #t (odd? (- n 1)))))
			  (odd?
			   (lambda (n) (if (zero? n) #f (even? (- n 1))))))
		   (even? 88)))
(define x 34)
(test 5 'let (let ((x 3)) (define x 5) x))
(test 34 'let x)
(test 6 'let (let () (define x 6) x))
(test 34 'let x)
(test 7 'let* (let* ((x 3)) (define x 7) x))
(test 34 'let* x)
(test 8 'let* (let* () (define x 8) x))
(test 34 'let* x)
(test 9 'letrec (letrec () (define x 9) x))
(test 34 'letrec x)
(test 10 'letrec (letrec ((x 3)) (define x 10) x))
(test 34 'letrec x)
(test 5 'letrec* (letrec* ((x 5)(y x)) y))
(define (error-test-let expr)
  (error-test (cons 'let expr))
  (error-test (cons 'let (cons 'name expr)))
  (error-test (cons 'let* expr))
  (error-test (cons 'letrec expr))
  (error-test (cons 'letrec* expr)))
(error-test-let 'x)
(error-test-let '(x))
(error-test-let '(()))
(error-test-let '(x ()))
(error-test '(let* x () 1))
(error-test '(letrec x () 1))
(error-test '(letrec* x () 1))
(error-test-let '(x . 1))
(error-test-let '(() . 1))
(error-test-let '(((x 1))))
(error-test-let '(((x 1)) . 1))
(error-test-let '(((x . 1)) 1))
(error-test-let '(((1 1)) 1))
(error-test-let '(((x 1) 1)  1))
(error-test-let '(((x 1) . 1)  1))
(error-test-let '(((x 1 1)) 1))
(SECTION 4 2 3)
(define x 0)
(test 6 'begin (begin (set! x 5) (+ x 1)))
(error-test '(begin . 1))
(error-test '(begin 1 . 2))
(SECTION 4 2 4)
(test '#(0 1 2 3 4) 'do (do ((vec (make-vector 5))
			    (i 0 (+ i 1)))
			   ((= i 5) vec)
			 (vector-set! vec i i)))
(test 25 'do (let ((x '(1 3 5 7 9)))
	       (do ((x x (cdr x))
		    (sum 0 (+ sum (car x))))
		   ((null? x) sum))))
(test 1 'let (let foo () 1))
(test '((6 1 3) (-5 -2)) 'let
      (let loop ((numbers '(3 -2 1 6 -5))
		 (nonneg '())
		 (neg '()))
	(cond ((null? numbers) (list nonneg neg))
	      ((negative? (car numbers))
	       (loop (cdr numbers)
		     nonneg
		     (cons (car numbers) neg)))
	      (else
	       (loop (cdr numbers)
		     (cons (car numbers) nonneg)
		     neg)))))
(error-test '(do))
(error-test '(do () ))
(error-test '(do () () ))
(error-test '(do (1) (#t 5) 5))
(error-test '(do ((1)) (#t 5) 5))
(error-test '(do ((x . 1)) (#t 5) 5))
(error-test '(do ((x 1) 2) (#t 5) 5))
(error-test '(do ((x 1) . 2) (#t 5) 5))
(error-test '(do ((x 1)) (#t . 5) 5))
(error-test '(do ((x 1)) (#t 5) . 5))
(SECTION 4 2 5)
(define f-check #t)
(define f (delay (begin (set! f-check #f) 5)))
(test #t (lambda () f-check))
(test 5 force f)
(test #f (lambda () f-check))
(error-test '(delay))
(error-test '(delay . 1))
(error-test '(delay 1 . 2))
(error-test '(delay 1 2))
(SECTION 4 2 6)
(test '(list 3 4) 'quasiquote `(list ,(+ 1 2) 4))
(test '(list a (quote a)) 'quasiquote (let ((name 'a)) `(list ,name ',name)))
(test '(a 3 4 5 6 b) 'quasiquote `(a ,(+ 1 2) ,@(map abs '(4 -5 6)) b))
(test '((foo 7) . cons)
	'quasiquote
	`((foo ,(- 10 3)) ,@(cdr '(c)) . ,(car '(cons))))

;;; sqt is defined here because not all implementations are required to
;;; support it. 
(define (sqt x)
	(do ((i 0 (+ i 1)))
	    ((> (* i i) x) (- i 1))))

(test '#(10 5 2 4 3 8) 'quasiquote `#(10 5 ,(sqt 4) ,@(map sqt '(16 9)) 8))
(test 5 'quasiquote `,(+ 2 3))
(test '(a `(b ,(+ 1 2) ,(foo 4 d) e) f)
      'quasiquote `(a `(b ,(+ 1 2) ,(foo ,(+ 1 3) d) e) f))
(test '(a `(b ,x ,'y d) e) 'quasiquote
	(let ((name1 'x) (name2 'y)) `(a `(b ,,name1 ,',name2 d) e)))
(test '(list 3 4) 'quasiquote (quasiquote (list (unquote (+ 1 2)) 4)))
(test '`(list ,(+ 1 2) 4) 'quasiquote '(quasiquote (list (unquote (+ 1 2)) 4)))
(error-test '(quasiquote))
(error-test '(quasiquote . 5))
(error-test '(quasiquote 1 . 2))
(error-test '(quasiquote 1 2))
(SECTION 5 2 1)
(define add3 (lambda (x) (+ x 3)))
(test 6 'define (add3 3))
(define first car)
(test 1 'define (first '(1 2)))
(error-test '(define))
(error-test '(define . x))
(error-test '(define x))
(error-test '(define x . 1))
(error-test '(define 1 2))
(error-test '(define (1) 1))
(error-test '(define (x 1) 1))
(error-test '(define x 1 . 2))
(error-test '(define x 1 2))
(SECTION 5 2 2)
(test 45 'define
	(let ((x 5))
		(define foo (lambda (y) (bar x y)))
		(define bar (lambda (a b) (+ (* a b) a)))
		(foo (+ x 3))))
(define x 34)
(define (foo) (define x 5) x)
(test 5 foo)
(test 34 'define x)
(define foo (lambda () (define x 5) x))
(test 5 foo)
(test 34 'define x)
(define (foo x) ((lambda () (define x 5) x)) x)
(test 88 foo 88)
(test 4 foo 4)
(test 34 'define x)

(defmacro define-alias (x y) `(define ,x ,y))
(test 45 'define
	(let ((x 5))
		(define-alias foo (lambda (y) (bar x y)))
		(define-alias bar (lambda (a b) (+ (* a b) a)))
		(foo (+ x 3))))
`(test 5 'define
      (let ()
	(define x 5)
	(define define (lambda (a b) (+ a b)))
	(define x 7)
	x))

'(error-test '(let ()
	       (define define 5)
	       (define y 6)
	       y))
(error-test '(let ()
	       (define x 5)))
(error-test '(let ()
	       (if #t
		   (define x 5))
	       5))
(SECTION 6 1)
(test #f not #t)
(test #f not 3)
(test #f not (list 3))
(test #t not #f)
(test #f not '())
(test #f not (list))
(test #f not 'nil)
(arity-test not 1 1)

(test #t boolean? #f)
(test #f boolean? 0)
(test #f boolean? '())
(arity-test boolean? 1 1)
(SECTION 6 2)
(test #t eqv? 'a 'a)
(test #f eqv? 'a 'b)
(test #t eqv? 2 2)
(test #t eqv? '() '())
(test #t eqv? '10000 '10000)
(test #f eqv? (cons 1 2)(cons 1 2))
(test #f eqv? (lambda () 1) (lambda () 2))
(test #f eqv? #f 'nil)
(let ((p (lambda (x) x)))
  (test #t eqv? p p))
(define gen-counter
 (lambda ()
   (let ((n 0))
      (lambda () (set! n (+ n 1)) n))))
(let ((g (gen-counter))) (test #t eqv? g g))
(test #f eqv? (gen-counter) (gen-counter))
(letrec ((f (lambda () (if (eqv? f g) 'f 'both)))
	 (g (lambda () (if (eqv? f g) 'g 'both))))
  (test #f eqv? f g))

(test #t eq? 'a 'a)
(test #f eq? (list 'a) (list 'a))
(test #t eq? '() '())
(test #t eq? car car)
(let ((x '(a))) (test #t eq? x x))
(let ((x '#())) (test #t eq? x x))
(let ((x (lambda (x) x))) (test #t eq? x x))

(test #t equal? 'a 'a)
(test #t equal? '(a) '(a))
(test #t equal? '(a (b) c) '(a (b) c))
(test #t equal? "abc" "abc")
(test #t equal? 2 2)
(test #t equal? (make-vector 5 'a) (make-vector 5 'a))

(arity-test eq? 2 2)
(arity-test eqv? 2 2)
(arity-test equal? 2 2)

(SECTION 6 3)
(test '(a b c d e) 'dot '(a . (b . (c . (d . (e . ()))))))
(define x (list 'a 'b 'c))
(define y x)
(and list? (test #t list? y))
(set-cdr! x 4)
(test '(a . 4) 'set-cdr! x)
(test #t eqv? x y)
(test '(a b c . d) 'dot '(a . (b . (c . d))))
(and list? (test #f list? y))
(and list? (let ((x (list 'a))) (set-cdr! x x) (test #f 'list? (list? x))))
(and list? (arity-test list? 1 1))

(test #t pair? '(a . b))
(test #t pair? '(a . 1))
(test #t pair? '(a b c))
(test #f pair? '())
(test #f pair? '#(a b))
(arity-test pair? 1 1)

(test '(a) cons 'a '())
(test '((a) b c d) cons '(a) '(b c d))
(test '("a" b c) cons "a" '(b c))
(test '(a . 3) cons 'a 3)
(test '((a b) . c) cons '(a b) 'c)
(arity-test cons 2 2) 

(test 'a car '(a b c))
(test '(a) car '((a) b c d))
(test 1 car '(1 . 2))
(arity-test car 1 1)
(error-test '(car 1))

(test '(b c d) cdr '((a) b c d))
(test 2 cdr '(1 . 2))
(arity-test cdr 1 1)
(error-test '(cdr 1))

(test '(a 7 c) list 'a (+ 3 4) 'c)
(test '() list)

(test 3 length '(a b c))
(test 3 length '(a (b) (c d e)))
(test 0 length '())
(arity-test length 1 1)
(error-test '(length 1))
(error-test '(length '(1 . 2)))

(define l '(1 2 3))
(set-cdr! l 5)
(test '(1 . 5) 'set-cdr! l)
(set-car! l 0)
(test '(0 . 5) 'set-car! l)
(arity-test set-car! 2 2)
(arity-test set-cdr! 2 2)
(error-test '(set-car! 4 4))
(error-test '(set-cdr! 4 4))

(test '(x y) append '(x) '(y))
(test '(a b c d) append '(a) '(b c d))
(test '(a (b) (c)) append '(a (b)) '((c)))
(test '() append)
(test '(a b c . d) append '(a b) '(c . d))
(test 'a append '() 'a)
(test 1 append 1)
(test '(1 . 2) append '(1) 2)
(test '(x y) append! '(x) '(y))
(test '(a b c d) append! '(a) '(b c d))
(test '(a (b) (c)) append! '(a (b)) '((c)))
(test '() append!)
(test '(a b c . d) append! '(a b) '(c . d))
(test 'a append! '() 'a)
(test 1 append! 1)
(test '(1 . 2) append '(1) 2)
(error-test '(append '(1 2 . 3) 1))
(error-test '(append '(1 2 3) 1 '(4 5 6)))
(error-test '(append! '(1 2 . 3) 1))
(error-test '(append! '(1 2 3) 1 '(4 5 6)))

(define l '(1 2))
(define l2 '(3 4 . 7))
(define l3 (append l l2))
(test '(1 2 3 4 . 7) 'append l3)
(set-car! l2 5)
(test '(1 2 5 4 . 7) 'append l3)
(set-car! l3 0)
(test '(0 2 5 4 . 7) 'append l3)
(test '(1 2) 'append l)

(define l '(1 2))
(define l2 '(3 4 . 7))
(define l3 (append! l l2))
(test '(1 2 3 4 . 7) 'append! l3)
(set-car! l2 5)
(test '(1 2 5 4 . 7) 'append! l3)
(set-car! l3 0)
(test '(0 2 5 4 . 7) 'append! l3)
(test '(0 2 5 4 . 7) 'append! l)

(test '(c b a) reverse '(a b c))
(test '((e (f)) d (b c) a) reverse '(a (b c) d (e (f))))
(arity-test reverse 1 1)
(error-test '(reverse 1))

(define l '(a b c))
(test '(c b a) reverse! l)
(test '(a) 'reverse! l)
(test '((e (f)) d (b c) a) reverse! '(a (b c) d (e (f))))
(arity-test reverse! 1 1)
(error-test '(reverse! 1))

(test 'c list-ref '(a b c d) 2)
(test 'c list-ref '(a b c d) 2.0)
(arity-test list-ref 2 2)
(error-test '(list-ref 1 1))
(error-test '(list-ref '(1) '(1)))
(error-test '(list-ref '(1) 1))
(error-test '(list-ref '(1) -1))

(test '(c d) list-tail '(a b c d) 2)
(test '(c d) list-tail '(a b c d) 2.0)
(arity-test list-tail 2 2)
(error-test '(list-tail 1 1))
(error-test '(list-tail '(1) '(1)))
(error-test '(list-tail '(1) 2))

(test '(a b c) memq 'a '(a b c))
(test '(b c) memq 'b '(a b c))
(test '#f memq 'a '(b c d))
(test '#f memq (list 'a) '(b (a) c))
(test '((a) c) member (list 'a) '(b (a) c))
(test '(101 102) memv 101 '(100 101 102))
(arity-test memq 2 2)
(arity-test member 2 2)
(arity-test memv 2 2)
(error-test '(memq 'a 1))
(error-test '(memq 'a '(1 . 2)))
(error-test '(memv 'a 1))
(error-test '(memv 'a '(1 . 2)))
(error-test '(member 'a 1))
(error-test '(member 'a '(1 . 2)))

(define e '((a 1) (b 2) (c 3)))
(test '(a 1) assq 'a e)
(test '(b 2) assq 'b e)
(test #f assq 'd e)
(test #f assq (list 'a) '(((a)) ((b)) ((c))))
(test '((a)) assoc (list 'a) '(((a)) ((b)) ((c))))
(test '(5 7) assv 5 '((2 3) (5 7) (11 13)))
(arity-test assq 2 2)
(arity-test assoc 2 2)
(arity-test assv 2 2)
(error-test '(assoc 1 1))
(error-test '(assq 1 1))
(error-test '(assv 1 1))
(error-test '(assoc 1 '(1 2)))
(error-test '(assq 1 '(1 2)))
(error-test '(assv 1 '(1 2)))
(error-test '(assoc 1 '((0) . 2)))
(error-test '(assq 1 '((0) . 2)))
(error-test '(assv 1 '((0) . 2)))

(SECTION 6 4)
(test #t symbol? 'foo)
(test #t symbol? (car '(a b)))
(test #f symbol? "bar")
(test #t symbol? 'nil)
(test #f symbol? '())
(test #f symbol? #f)
;;; But first, what case are symbols in?  Determine the standard case:
(define char-standard-case char-upcase)
(if (string=? (symbol->string 'A) "a")
    (set! char-standard-case char-downcase))
(test #t 'standard-case
      (string=? (symbol->string 'a) (symbol->string 'A)))
(test #t 'standard-case
      (or (string=? (symbol->string 'a) "A")
	  (string=? (symbol->string 'A) "a")))
(define (str-copy s)
  (let ((v (make-string (string-length s))))
    (do ((i (- (string-length v) 1) (- i 1)))
	((< i 0) v)
      (string-set! v i (string-ref s i)))))
(define (string-standard-case s)
  (set! s (str-copy s))
  (do ((i 0 (+ 1 i))
       (sl (string-length s)))
      ((>= i sl) s)
      (string-set! s i (char-standard-case (string-ref s i)))))
(test (string-standard-case "flying-fish") symbol->string 'flying-fish)
(test (string-standard-case "martin") symbol->string 'Martin)
(test "Malvina" symbol->string (string->symbol "Malvina"))
(test #t 'standard-case (eq? 'a 'A))

(define x (string #\a #\b))
(define y (string->symbol x))
(string-set! x 0 #\c)
(test "cb" 'string-set! x)
(test "ab" symbol->string y)
(test y string->symbol "ab")

(test #t eq? 'mISSISSIppi 'mississippi)
(test #f 'string->symbol (eq? 'bitBlt (string->symbol "bitBlt")))
(test 'JollyWog string->symbol (symbol->string 'JollyWog))

(arity-test symbol? 1 1)

(SECTION 6 5 5)
(test #t number? 3)
(test #t complex? 3)
(test #t real? 3)
(test #t rational? 3)
(test #t integer? 3)

(test #t exact? 3)

(arity-test inexact? 1 1)
(arity-test number? 1 1)
(arity-test complex? 1 1)
(arity-test real? 1 1)
(arity-test rational? 1 1)
(arity-test integer? 1 1)
(arity-test exact? 1 1)
(arity-test inexact? 1 1)

(test #t = 0)
(test #t > 0)
(test #t < 0)
(test #t >= 0)
(test #t <= 0)
(test #t = 22 22 22)
(test #t = 22 22)
(test #f = 34 34 35)
(test #f = 34 35)
(test #t > 3 -6246)
(test #f > 9 9 -2424)
(test #t >= 3 -4 -6246)
(test #t >= 9 9)
(test #f >= 8 9)
(test #t < -1 2 3 4 5 6 7 8)
(test #f < -1 2 3 4 4 5 6 7)
(test #t <= -1 2 3 4 5 6 7 8)
(test #t <= -1 2 3 4 4 5 6 7)
(test #f < 1 3 2)
(test #f >= 1 3 2)

(error-test '(= 1 'a))
(error-test '(= 'a 1))
(error-test '(= 'a))
(error-test '(> 1 'a))
(error-test '(> 'a 1))
(error-test '(< 1 'a))
(error-test '(< 'a 1))
(error-test '(>= 1 'a))
(error-test '(>= 'a 1))
(error-test '(<= 1 'a))
(error-test '(<= 'a 1))

(test #t zero? 0)
(test #f zero? 1)
(test #f zero? -1)
(test #f zero? -100)
(test #t positive? 4)
(test #f positive? -4)
(test #f positive? 0)
(test #f negative? 4)
(test #t negative? -4)
(test #f negative? 0)
(test #t odd? 3)
(test #f odd? 2)
(test #f odd? -4)
(test #t odd? -1)
(test #f even? 3)
(test #t even? 2)
(test #t even? -4)
(test #f even? -1)

(arity-test zero? 1 1)
(arity-test positive? 1 1)
(arity-test negative? 1 1)
(arity-test odd? 1 1)
(arity-test even? 1 1)

(test 5 max 5)
(test 5 min 5)
(test 38 max 34 5 7 38 6)
(test -24 min 3  5 5 330 4 -24)
(test 38.0 max 34 5.0 7 38 6)
(test -24.0 min 3  5 5 330 4 -24.0)

(error-test '(max 0 'a))
(error-test '(min 0 'a))
(error-test '(max 'a 0))
(error-test '(min 'a 0))
(error-test '(max 'a))
(error-test '(min 'a))

(test 0 +)
(test 7 + 3 4)
(test 7.0 + 3 4.0)
(test '3 + 3)
(test 0 +)
(test 4 * 4)
(test 16.0 * 4 4.0)
(test 1 *)

(error-test '(* 'a 0))
(error-test '(+ 'a 0))
(error-test '(+ 0 'a))
(error-test '(* 0 'a))
(error-test '(+ 'a))
(error-test '(* 'a))

(test -1 - 3 4)
(test -3 - 3)
(test -1.0 - 3.0 4)
(test -3.0 - 3.0)
(test 7 abs -7)
(test 7.0 abs -7.0)
(test 7 abs 7)
(test 0 abs 0)

(arity-test abs 1 1)
(error-test '(-))
(error-test '(abs 'a))

(test 5 quotient 35 7)
(test 5.0 quotient 35 7.0)
(test -5 quotient -35 7)
(test -5.0 quotient -35 7.0)
(test -5 quotient 35 -7)
(test -5.0 quotient 35 -7.0)
(test 5 quotient -35 -7)
(test 5.0 quotient -35 -7.0)
(test 1 modulo 13 4)
(test 1 remainder 13 4)
(test 1.0 modulo 13 4.0)
(test 1.0 remainder 13 4.0)
(test 3 modulo -13 4)
(test -1 remainder -13 4)
(test 3.0 modulo -13 4.0)
(test -1.0 remainder -13 4.0)
(test -3 modulo 13 -4)
(test 1 remainder 13 -4)
(test -3.0 modulo 13.0 -4)
(test 1.0 remainder 13.0 -4)
(test -1 modulo -13 -4)
(test -1 remainder -13 -4)
(test -1.0 modulo -13 -4.0)
(test -1.0 remainder -13 -4.0)
(define (divtest n1 n2)
	(= n1 (+ (* n2 (quotient n1 n2))
		 (remainder n1 n2))))
(test #t divtest 238 9)
(test #t divtest -238 9)
(test #t divtest 238 -9)
(test #t divtest -238 -9)

(arity-test quotient 2 2)
(arity-test modulo 2 2)
(arity-test remainder 2 2)

(error-test '(quotient 'a 1))
(error-test '(quotient 1 'a))
(error-test '(modulo 'a 1))
(error-test '(modulo 1 'a))
(error-test '(remainder 'a 1))
(error-test '(remainder 1 'a))
(error-test '(quotient 'a 1.0))
(error-test '(quotient 1.0 'a))
(error-test '(modulo 'a 1.0))
(error-test '(modulo 1.0 'a))
(error-test '(remainder 'a 1.0))
(error-test '(remainder 1.0 'a))

(test 4 gcd 0 4)
(test 4 gcd -4 0)
(test 4 gcd 32 -36)
(test 0 gcd)
(test 5 gcd 5)
(test 5.0 gcd 5.0 10)
(test 288 lcm 32 -36)
(test 1 lcm)
(test 5 lcm 5)
(test 30.0 lcm 5 6.0)

(error-test '(gcd 'a))
(error-test '(gcd 'a 1))
(error-test '(gcd 1 'a))
(error-test '(lcm 'a))
(error-test '(lcm 'a 1))
(error-test '(lcm 1 'a))

;;;;From: fred@sce.carleton.ca (Fred J Kaudel)
;;; Modified by jaffer.
(define (test-inexact)
  (define f3.9 (string->number "3.9"))
  (define f4.0 (string->number "4.0"))
  (define f-3.25 (string->number "-3.25"))
  (define f.25 (string->number ".25"))
  (define f4.5 (string->number "4.5"))
  (define f3.5 (string->number "3.5"))
  (define f0.0 (string->number "0.0"))
  (define f0.8 (string->number "0.8"))
  (define f1.0 (string->number "1.0"))
  (define wto write-test-obj)
  (define dto display-test-obj)
  (define lto load-test-obj)
  (newline)
  (display ";testing inexact numbers; ")
  (newline)
  (SECTION 6 5 5)
  (test #t inexact? f3.9)
  (test #t 'inexact? (inexact? (max f3.9 4)))
  (test f4.0 'max (max f3.9 4))
  (test f4.0 'exact->inexact (exact->inexact 4))
  (test (- f4.0) round (- f4.5))
  (test (- f4.0) round (- f3.5))
  (test (- f4.0) round (- f3.9))
  (test f0.0 round f0.0)
  (test f0.0 round f.25)
  (test f1.0 round f0.8)
  (test f4.0 round f3.5)
  (test f4.0 round f4.5)
  (set! write-test-obj (list f.25 f-3.25));.25 inexact errors less likely.
  (set! display-test-obj (list f.25 f-3.25));.3 often has such errors (~10^-13)
  (set! load-test-obj (list 'define 'foo (list 'quote write-test-obj)))
  (test #t call-with-output-file
      "tmp3"
      (lambda (test-file)
	(write-char #\; test-file)
	(display write-test-obj test-file)
	(newline test-file)
	(write load-test-obj test-file)
	(output-port? test-file))
      'truncate)
  (check-test-file "tmp3")
  (set! write-test-obj wto)
  (set! display-test-obj dto)
  (set! load-test-obj lto)
  (let ((x (string->number "4195835.0"))
	(y (string->number "3145727.0")))
    (test #t 'pentium-fdiv-bug (> f1.0 (- x (* (/ x y) y)))))
  '(report-errs))

(define (test-bignum)
  (define tb
    (lambda (n1 n2)
      (= n1 (+ (* n2 (quotient n1 n2))
	       (remainder n1 n2)))))
  (newline)
  (display ";testing bignums; ")
  (newline)
  (section 6 5 5)
  (test 0 modulo -2177452800 86400)
  (test 0 modulo 2177452800 -86400)
  (test 0 modulo 2177452800 86400)
  (test 0 modulo -2177452800 -86400)

  (test 86399 modulo -2177452801 86400)
  (test -1 modulo 2177452799 -86400)
  (test 1 modulo 2177452801 86400)
  (test -86399 modulo -2177452799 -86400)

  (test #t 'remainder (tb 281474976710655 65535))
  (test #t 'remainder (tb 281474976710654 65535))
  (SECTION 6 5 6)
  (test 281474976710655 string->number "281474976710655")
  (test "281474976710655" number->string 281474976710655)
  '(report-errs))


(SECTION 6 5 6)
(test "0" number->string 0)
(test "100" number->string 100)
(test "100" number->string 256 16)
(test 100 string->number "100")
(test 256 string->number "100" 16)
(test #f string->number "")
(test #f string->number ".")
(test #f string->number "d")
(test #f string->number "D")
(test #f string->number "i")
(test #f string->number "I")
(test #f string->number "3i")
(test #f string->number "3I")
(test #f string->number "33i")
(test #f string->number "33I")
(test #f string->number "3.3i")
(test #f string->number "3.3I")
(test #f string->number "-")
(test #f string->number "+")
(arity-test string->number 1 1)
(error-test '(string->number 'a))
(SECTION 6 6)
(test #t eqv? '#\  #\Space)
(test #t eqv? #\space '#\Space)
(test #t char? #\a)
(test #t char? #\()
(test #t char? #\ )
(test #t char? '#\newline)
(arity-test char? 1 1)

(test #f char=? #\A #\B)
(test #f char=? #\a #\b)
(test #f char=? #\9 #\0)
(test #t char=? #\A #\A)
(arity-test char=? 2 2)
(error-test '(char=? #\a 1)) 
(error-test '(char=? 1 #\a))

(test #t char<? #\A #\B)
(test #t char<? #\a #\b)
(test #f char<? #\9 #\0)
(test #f char<? #\A #\A)
(arity-test char<? 2 2)
(error-test '(char<? #\a 1)) 
(error-test '(char<? 1 #\a))

(test #f char>? #\A #\B)
(test #f char>? #\a #\b)
(test #t char>? #\9 #\0)
(test #f char>? #\A #\A)
(arity-test char>? 2 2)
(error-test '(char>? #\a 1)) 
(error-test '(char>? 1 #\a))

(test #t char<=? #\A #\B)
(test #t char<=? #\a #\b)
(test #f char<=? #\9 #\0)
(test #t char<=? #\A #\A)
(arity-test char<=? 2 2)
(error-test '(char<=? #\a 1)) 
(error-test '(char<=? 1 #\a))

(test #f char>=? #\A #\B)
(test #f char>=? #\a #\b)
(test #t char>=? #\9 #\0)
(test #t char>=? #\A #\A)
(arity-test char>=? 2 2)
(error-test '(char>=? #\a 1)) 
(error-test '(char>=? 1 #\a))

(test #f char-ci=? #\A #\B)
(test #f char-ci=? #\a #\B)
(test #f char-ci=? #\A #\b)
(test #f char-ci=? #\a #\b)
(test #f char-ci=? #\9 #\0)
(test #t char-ci=? #\A #\A)
(test #t char-ci=? #\A #\a)
(arity-test char-ci=? 2 2)
(error-test '(char-ci=? #\a 1)) 
(error-test '(char-ci=? 1 #\a))

(test #t char-ci<? #\A #\B)
(test #t char-ci<? #\a #\B)
(test #t char-ci<? #\A #\b)
(test #t char-ci<? #\a #\b)
(test #f char-ci<? #\9 #\0)
(test #f char-ci<? #\A #\A)
(test #f char-ci<? #\A #\a)
(arity-test char-ci<? 2 2)
(error-test '(char-ci<? #\a 1)) 
(error-test '(char-ci<? 1 #\a))

(test #f char-ci>? #\A #\B)
(test #f char-ci>? #\a #\B)
(test #f char-ci>? #\A #\b)
(test #f char-ci>? #\a #\b)
(test #t char-ci>? #\9 #\0)
(test #f char-ci>? #\A #\A)
(test #f char-ci>? #\A #\a)
(arity-test char-ci>? 2 2)
(error-test '(char-ci>? #\a 1)) 
(error-test '(char-ci>? 1 #\a))

(test #t char-ci<=? #\A #\B)
(test #t char-ci<=? #\a #\B)
(test #t char-ci<=? #\A #\b)
(test #t char-ci<=? #\a #\b)
(test #f char-ci<=? #\9 #\0)
(test #t char-ci<=? #\A #\A)
(test #t char-ci<=? #\A #\a)
(arity-test char-ci<=? 2 2)
(error-test '(char-ci<=? #\a 1)) 
(error-test '(char-ci<=? 1 #\a))

(test #f char-ci>=? #\A #\B)
(test #f char-ci>=? #\a #\B)
(test #f char-ci>=? #\A #\b)
(test #f char-ci>=? #\a #\b)
(test #t char-ci>=? #\9 #\0)
(test #t char-ci>=? #\A #\A)
(test #t char-ci>=? #\A #\a)
(arity-test char-ci>=? 2 2)
(error-test '(char-ci>=? #\a 1)) 
(error-test '(char-ci>=? 1 #\a))

(test #t char-alphabetic? #\a)
(test #t char-alphabetic? #\A)
(test #t char-alphabetic? #\z)
(test #t char-alphabetic? #\Z)
(test #f char-alphabetic? #\0)
(test #f char-alphabetic? #\9)
(test #f char-alphabetic? #\space)
(test #f char-alphabetic? #\;)
(arity-test char-alphabetic? 1 1)
(error-test '(char-alphabetic? 1))

(test #f char-numeric? #\a)
(test #f char-numeric? #\A)
(test #f char-numeric? #\z)
(test #f char-numeric? #\Z)
(test #t char-numeric? #\0)
(test #t char-numeric? #\9)
(test #f char-numeric? #\space)
(test #f char-numeric? #\;)
(arity-test char-numeric? 1 1)
(error-test '(char-numeric? 1))

(test #f char-whitespace? #\a)
(test #f char-whitespace? #\A)
(test #f char-whitespace? #\z)
(test #f char-whitespace? #\Z)
(test #f char-whitespace? #\0)
(test #f char-whitespace? #\9)
(test #t char-whitespace? #\space)
(test #f char-whitespace? #\;)
(arity-test char-whitespace? 1 1)
(error-test '(char-whitespace? 1))

(test #f char-upper-case? #\0)
(test #f char-upper-case? #\9)
(test #f char-upper-case? #\space)
(test #f char-upper-case? #\;)
(arity-test char-upper-case? 1 1)
(error-test '(char-upper-case? 1))

(test #f char-lower-case? #\0)
(test #f char-lower-case? #\9)
(test #f char-lower-case? #\space)
(test #f char-lower-case? #\;)
(arity-test char-lower-case? 1 1)
(error-test '(char-lower-case? 1))

(test #\. integer->char (char->integer #\.))
(test #\A integer->char (char->integer #\A))
(test #\a integer->char (char->integer #\a))
(arity-test integer->char 1 1)
(arity-test char->integer 1 1)
(error-test '(integer->char 5.0))
(error-test '(integer->char 'a))
(error-test '(char->integer 5))

(test #\A char-upcase #\A)
(test #\A char-upcase #\a)
(test #\a char-downcase #\A)
(test #\a char-downcase #\a)
(arity-test char-upcase 1 1)
(arity-test char-downcase 1 1)
(error-test '(char-upcase 2))
(error-test '(char-downcase 2))

(SECTION 6 7)
(test #t string? "The word \"recursion\\\" has many meanings.")
(test #t string? "")
(arity-test string? 1 1)
(test 3 'make-string (string-length (make-string 3)))
(test 3 'make-string (string-length (make-string 3.0)))
(arity-test make-string 1 2)
(error-test '(make-string "hello"))
(error-test '(make-string 5 "hello"))
(error-test '(make-string -5 #\f))

(define f (make-string 3 #\*))
(test "?**" 'string-set! (begin (string-set! f 0 #\?) f))
(define f (make-string 3.0 #\*))
(test "*?*" 'string-set! (begin (string-set! f 1.0 #\?) f))
(arity-test string-set! 3 3)
(error-test '(string-set! 'a "hello" #\a))
(error-test '(string-set! 4 'hello #\a))
(error-test '(string-set! 4 "hello" 'a))
(error-test '(string-set! 5 "hello" #\a))
(error-test '(string-set! -1 "hello" #\a))
(test "abc" string #\a #\b #\c)
(test "" string)
(error-test '(string #\a 1))
(error-test '(string 1 #\a))
(error-test '(string 1))
(test 3 string-length "abc")
(test 0 string-length "")
(arity-test string-length 1 1)
(error-test '(string-length 'apple))
(test #\a string-ref "abc" 0)
(test #\c string-ref "abc" 2)
(test #\c string-ref "abc" 2.0)
(arity-test string-ref 2 2)
(error-test '(string-ref 'apple 4))
(error-test '(string-ref "apple" '(4)))
(error-test '(string-ref "apple" 5))
(error-test '(string-ref "apple" -1))
(test "" substring "ab" 0 0)
(test "" substring "ab" 1 1)
(test "" substring "ab" 2 2)
(test "" substring "ab" 2.0 2.0)
(test "a" substring "ab" 0 1)
(test "b" substring "ab" 1 2)
(test "ab" substring "ab" 0 2)
(arity-test substring 3 3)
(error-test '(substring 'hello 2 3))
(error-test '(substring "hello" "2" 3))
(error-test '(substring "hello" 2 "3"))
(error-test '(substring "hello" 2 7))
(error-test '(substring "hello" -2 3))
(test "foobar" string-append "foo" "bar")
(test "foo" string-append "foo")
(test "foo" string-append "foo" "")
(test "foo" string-append "" "foo")
(test "" string-append)
(error-test '(string-append 1))
(error-test '(string-append "hello" 1))
(error-test '(string-append "hello" 1 "done"))
(test "" make-string 0)
(define s "hello")
(define s2 (string-copy s))
(test "hello" 'string-copy s2)
(string-set! s 2 #\x)
(test "hello" 'string-copy s2)
(string-fill! s #\x)
(test "xxxxx" 'string-fill! s)
(arity-test string-copy 1 1)
(arity-test string-fill! 2 2)
(error-test '(string-copy 'blah))
(error-test '(string-fill! "oops" 5))

(test #t string=? "" "")
(test #f string<? "" "")
(test #f string>? "" "")
(test #t string<=? "" "")
(test #t string>=? "" "")
(test #t string-ci=? "" "")
(test #f string-ci<? "" "")
(test #f string-ci>? "" "")
(test #t string-ci<=? "" "")
(test #t string-ci>=? "" "")

(test #f string=? "A" "B")
(test #f string=? "a" "b")
(test #f string=? "9" "0")
(test #t string=? "A" "A")

(test #t string<? "A" "B")
(test #t string<? "a" "b")
(test #f string<? "9" "0")
(test #f string<? "A" "A")

(test #f string>? "A" "B")
(test #f string>? "a" "b")
(test #t string>? "9" "0")
(test #f string>? "A" "A")

(test #t string<=? "A" "B")
(test #t string<=? "a" "b")
(test #f string<=? "9" "0")
(test #t string<=? "A" "A")

(test #f string>=? "A" "B")
(test #f string>=? "a" "b")
(test #t string>=? "9" "0")
(test #t string>=? "A" "A")

(test #f string-ci=? "A" "B")
(test #f string-ci=? "a" "B")
(test #f string-ci=? "A" "b")
(test #f string-ci=? "a" "b")
(test #f string-ci=? "9" "0")
(test #t string-ci=? "A" "A")
(test #t string-ci=? "A" "a")

(test #t string-ci<? "A" "B")
(test #t string-ci<? "a" "B")
(test #t string-ci<? "A" "b")
(test #t string-ci<? "a" "b")
(test #f string-ci<? "9" "0")
(test #f string-ci<? "A" "A")
(test #f string-ci<? "A" "a")

(test #f string-ci>? "A" "B")
(test #f string-ci>? "a" "B")
(test #f string-ci>? "A" "b")
(test #f string-ci>? "a" "b")
(test #t string-ci>? "9" "0")
(test #f string-ci>? "A" "A")
(test #f string-ci>? "A" "a")

(test #t string-ci<=? "A" "B")
(test #t string-ci<=? "a" "B")
(test #t string-ci<=? "A" "b")
(test #t string-ci<=? "a" "b")
(test #f string-ci<=? "9" "0")
(test #t string-ci<=? "A" "A")
(test #t string-ci<=? "A" "a")

(test #f string-ci>=? "A" "B")
(test #f string-ci>=? "a" "B")
(test #f string-ci>=? "A" "b")
(test #f string-ci>=? "a" "b")
(test #t string-ci>=? "9" "0")
(test #t string-ci>=? "A" "A")
(test #t string-ci>=? "A" "a")

(map (lambda (pred)
       (arity-test pred 2 2)
       (error-test `(,pred "a" 1))
       (error-test `(,pred 1 "a")))
     (list string=? 
	   string>? 
	   string<? 
	   string>=? 
	   string<=? 
	   string-ci=? 
	   string-ci>? 
	   string-ci<? 
	   string-ci>=? 
	   string-ci<=?))

(SECTION 6 8)
(test #t vector? '#(0 (2 2 2 2) "Anna"))
(test #t vector? '#())
(arity-test vector? 1 1)
(test '#(a b c) vector 'a 'b 'c)
(test '#() vector)
(test 3 vector-length '#(0 (2 2 2 2) "Anna"))
(test 0 vector-length '#())
(arity-test vector-length 1 1)
(error-test '(vector-length "apple"))
(test 8 vector-ref '#(1 1 2 3 5 8 13 21) 5)
(test 8 vector-ref '#(1 1 2 3 5 8 13 21) 5.0)
(arity-test vector-ref 2 2)
(error-test '(vector-ref "apple" 3))
(error-test '(vector-ref #(4 5 6) 3))
(error-test '(vector-ref #(4 5 6) -1))
(error-test '(vector-ref #(4 5 6) "2"))
(test '#(0 ("Sue" "Sue") "Anna") 'vector-set
	(let ((vec (vector 0 '(2 2 2 2) "Anna")))
	  (vector-set! vec 1 '("Sue" "Sue"))
	  vec))
(test '#(hi hi) make-vector 2 'hi)
(test '#() make-vector 0)
(test '#() make-vector 0 'a)
(test '#(a) make-vector 1.0 'a)
(arity-test make-vector 1 2)
(error-test '(make-vector "a" 'a))
(error-test '(make-vector -1 'a))
(arity-test vector-set! 3 3)
(error-test '(vector-set! #(1 2 3) -1 'x))
(error-test '(vector-set! #(1 2 3) 3 'x))
(error-test '(vector-set! '(1 2 3) 2 'x))
(error-test '(vector-set! #(1 2 3) "2" 'x))
(define v #(1 2 3))
(vector-fill! v 0)
(test #(0 0 0) 'vector-fill! v)
(arity-test vector-fill! 2 2)
(error-test '(vector-fill! '(1 2 3) 0))

(SECTION 6 9)
(test #t procedure? car)
(test #f procedure? 'car)
(test #t procedure? (lambda (x) (* x x)))
(test #f procedure? '(lambda (x) (* x x)))
(test #t call-with-current-continuation procedure?)
(test #t call-with-escaping-continuation procedure?)
(test #t procedure? (case-lambda ((x) x) ((x y) (+ x y))))
(arity-test procedure? 1 1)

(test 7 apply + (list 3 4))
(test 7 apply (lambda (a b) (+ a b)) (list 3 4))
(test 17 apply + 10 (list 3 4))
(test '() apply list '())
(define compose (lambda (f g) (lambda args (f (apply g args)))))
(test 30 (compose sqt *) 12 75)
(error-test '(apply))
(error-test '(apply (lambda x x)))
(error-test '(apply (lambda x x) 1))
(error-test '(apply (lambda x x) 1 2))
(error-test '(apply (lambda x x) 1 '(2 . 3)))

(test '(b e h) map cadr '((a b) (d e) (g h)))
(test '(5 7 9) map + '(1 2 3) '(4 5 6))
(test '#(0 1 4 9 16) 'for-each
	(let ((v (make-vector 5)))
		(for-each (lambda (i) (vector-set! v i (* i i)))
			'(0 1 2 3 4))
		v))

(define (map-tests map)
  (error-test `(,map (lambda (x y) (+ x y)) '(1 2) '1))
  (error-test `(,map (lambda (x y) (+ x y)) '2 '(1 2)))
  (error-test `(,map (lambda (x y) (+ x y)) '(1 2) '(1 2 3)))
  (error-test `(,map (lambda (x y) (+ x y)) '(1 2 3) '(1 2)))
  (error-test `(,map (lambda (x y) (+ x y)) '(1 2 . 3) '(1 2)))
  (error-test `(,map (lambda (x y) (+ x y)) '(1 2 . 3) '(1 2 3)))
  (error-test `(,map (lambda (x y) (+ x y)) '(1 2) '(1 2 . 3)))
  (error-test `(,map (lambda (x y) (+ x y)) '(1 2 3) '(1 2 . 3)))
  (error-test `(,map))
  (error-test `(,map (lambda (x y) (+ x y)))))
(map-tests map)
(map-tests for-each)

(test -3 call-with-current-continuation
		(lambda (exit)
		 (for-each (lambda (x) (if (negative? x) (exit x)))
		 	'(54 0 37 -3 245 19))
		#t))
(define list-length
 (lambda (obj)
  (call-with-current-continuation
   (lambda (return)
    (letrec ((r (lambda (obj) (cond ((null? obj) 0)
				((pair? obj) (+ (r (cdr obj)) 1))
				(else (return #f))))))
	(r obj))))))
(test 4 list-length '(1 2 3 4))
(test #f list-length '(a b . c))
(test '() map cadr '())

;;; This tests full conformance of call-with-current-continuation.  It
;;; is a separate test because some schemes do not support call/cc
;;; other than escape procedures.  I am indebted to
;;; raja@copper.ucs.indiana.edu (Raja Sooriamurthi) for fixing this
;;; code.  The function leaf-eq? compares the leaves of 2 arbitrary
;;; trees constructed of conses.  
(define (next-leaf-generator obj eot)
  (letrec ((return #f)
	   (cont (lambda (x)
		   (recurx obj)
		   (set! cont (lambda (x) (return eot)))
		   (cont #f)))
	   (recurx (lambda (obj)
		      (if (pair? obj)
			  (for-each recurx obj)
			  (call-with-current-continuation
			   (lambda (c)
			     (set! cont c)
			     (return obj)))))))
    (lambda () (call-with-current-continuation
		(lambda (ret) (set! return ret) (cont #f))))))
(define (leaf-eq? x y)
  (let* ((eot (list 'eot))
	 (xf (next-leaf-generator x eot))
	 (yf (next-leaf-generator y eot)))
    (letrec ((loop (lambda (x y)
		     (cond ((not (eq? x y)) #f)
			   ((eq? eot x) #t)
			   (else (loop (xf) (yf)))))))
      (loop (xf) (yf)))))
(define (test-cont)
  (newline)
  (display ";testing continuations; ")
  (SECTION 6 9)
  (test #t leaf-eq? '(a (b (c))) '((a) b c))
  (test #f leaf-eq? '(a (b (c))) '((a) b c d))
  '(report-errs))

(arity-test call/cc 1 1)
(error-test '(call/cc 4))

(SECTION 6 10 1)
(test #t input-port? (current-input-port))
(test #t output-port? (current-output-port))
(test #t call-with-input-file "test.scm" input-port?)
(define this-file (open-input-file "test.scm"))
(test #t input-port? this-file)
(arity-test input-port? 1 1)
(arity-test output-port? 1 1)
(SECTION 6 10 2)
(test #\; peek-char this-file)
(arity-test peek-char 0 1)
(test #\; read-char this-file)
(arity-test read-char 0 1)
(test '(define cur-section '()) read this-file)
(arity-test read 0 1)
(test #\( peek-char this-file)
(test '(define errs '()) read this-file)
(close-input-port this-file)
(close-input-port this-file)
(arity-test close-input-port 1 1)
(arity-test close-output-port 1 1)
(error-test '(peek-char 5))
(error-test '(read-char 5))
(error-test '(read 5))
(error-test '(close-input-port 5))
(error-test '(close-output-port 5))
(define (check-test-file name)
  (define test-file (open-input-file name))
  (test #t 'input-port?
	(call-with-input-file
	    name
	  (lambda (test-file)
	    (test load-test-obj read test-file)
	    (test #t eof-object? (peek-char test-file))
	    (test #t eof-object? (read-char test-file))
	    (input-port? test-file))))
  (test #\; read-char test-file)
  (test display-test-obj read test-file)
  (test load-test-obj read test-file)
  (close-input-port test-file))
(SECTION 6 10 3)
(define write-test-obj
  '(#t #f #\a () 9739 -3 . #((test) "te \" \" st" "" test #() b c)))
(define display-test-obj
  '(#t #f a () 9739 -3 . #((test) te " " st test #() b c)))
(define load-test-obj
  (list 'define 'foo (list 'quote write-test-obj)))
(test #t call-with-output-file
      "tmp1"
      (lambda (test-file)
	(write-char #\; test-file)
	(display write-test-obj test-file)
	(newline test-file)
	(write load-test-obj test-file)
	(output-port? test-file))
      `truncate)
(check-test-file "tmp1")

(define test-file (open-output-file "tmp2" 'truncate))
(write-char #\; test-file)
(display write-test-obj test-file)
(newline test-file)
(write load-test-obj test-file)
(test #t output-port? test-file)
(close-output-port test-file)
(check-test-file "tmp2")
(define (test-sc4)
  (newline)
  (display ";testing scheme 4 functions; ")
  (SECTION 6 7)
  (test '(#\P #\space #\l) string->list "P l")
  (test '() string->list "")
  (test "1\\\"" list->string '(#\1 #\\ #\"))
  (test "" list->string '())
  (arity-test list->string 1 1)
  (arity-test string->list 1 1)
  (error-test '(string->list 'hello))
  (error-test '(list->string 'hello))
  (error-test '(list->string '(#\h . #\e)))
  (SECTION 6 8)
  (test '(dah dah didah) vector->list '#(dah dah didah))
  (test '() vector->list '#())
  (test '#(dididit dah) list->vector '(dididit dah))
  (test '#() list->vector '())
  (arity-test list->vector 1 1)
  (arity-test vector->list 1 1)
  (error-test '(vector->list 'hello))
  (error-test '(list->vector 'hello))
  (error-test '(list->vector '(#\h . #\e)))
  (SECTION 6 10 4)
  (load "tmp1")
  (test write-test-obj 'load foo)
  (report-errs))

(test-sc4)
(test-inexact)
(test-bignum)
(test-cont)

(report-errs)

"last item in file"
